import json
import re
import typing

import numpy as np

from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op, weave_class


@op(name="root-string")
def string(v: str) -> str:
    return v


@op(name="string-lastLetter")
def lastLetter(v: str) -> str:
    return v[-1]


def _json_parse(string: str) -> typing.Any:
    return json.loads(string)


@weave_class(weave_type=types.String)
class String:
    @op(
        name="string-_set",
        input_type={"self": types.String(), "val": types.String()},
        output_type=types.String(),
    )
    def set(self, val):
        return val

    @op(name="string-equal")
    def __eq__(lhs: typing.Optional[str], rhs: typing.Optional[str]) -> bool:  # type: ignore[misc]
        return lhs == rhs

    @op(name="string-notEqual")
    def __ne__(lhs: typing.Optional[str], rhs: typing.Optional[str]) -> bool:  # type: ignore[misc]
        return lhs != rhs

    @op(name="string-contains")
    def __contains__(str: str, sub: str) -> bool:  # type: ignore[misc]
        return sub in str

    @op(name="string-in")
    def in_(lhs: str, rhs: typing.Optional[str]) -> bool:  # type: ignore[misc]
        rhs = rhs or ""
        return lhs in rhs

    @op(name="string-add")
    def __add__(lhs: str, rhs: typing.Optional[str]) -> str:  # type: ignore[misc]
        if rhs == None:
            return None  # type: ignore
        return lhs + rhs  # type: ignore

    @op(name="string-len")
    def len(str: str) -> int:  # type: ignore[misc]
        return len(str)

    # This should be string __add__, which is implemented above, so as not
    # to list append which is a mutation
    @op(name="string-append")
    def append(str: str, suffix: typing.Optional[str]) -> str:  # type: ignore[misc]
        suffix = suffix or ""
        return str + suffix

    @op(name="string-prepend")
    def prepend(str: str, prefix: typing.Optional[str]) -> str:  # type: ignore[misc]
        prefix = prefix or ""
        return prefix + str

    @op(name="string-split")
    def split(str: str, sep: typing.Optional[str]) -> list[str]:  # type: ignore[misc]
        if sep is None:
            return list(str)
        return str.split(sep)

    @op(name="string-partition")
    def partition(str: str, sep: typing.Optional[str]) -> list[str]:  # type: ignore[misc]
        if sep is None:
            return [str, "", ""]
        return list(str.partition(sep))

    @op(name="string-startsWith")
    def startswith(str: str, prefix: typing.Optional[str]) -> bool:  # type: ignore[misc]
        prefix = prefix or ""
        return str.startswith(prefix)

    @op(name="string-endsWith")
    def endswith(str: str, suffix: typing.Optional[str]) -> bool:  # type: ignore[misc]
        suffix = suffix or ""
        return str.endswith(suffix)

    @op(name="string-isAlpha")
    def isalpha(str: str) -> bool:  # type: ignore[misc]
        return str.isalpha()

    @op(name="string-isNumeric")
    def isnumeric(str: str) -> bool:  # type: ignore[misc]
        # custom logic to match JS
        if str == "":
            return False
        if str[0] == "-":
            str = str[1:]
        if str == "":
            return False
        if str[0] == ".":
            str = str[1:]
            if "." in str:
                return False
        if str == "":
            return False
        if str[-1] == ".":
            str = str[:-1]
            if "." in str:
                return False
        if str == "":
            return False
        parts = str.split(".")
        if len(parts) > 2:
            return False
        for part in parts:
            if not part.isnumeric():
                return False
        return True

    @op(name="string-isAlnum")
    def isalnum(str: str) -> bool:  # type: ignore[misc]
        return str.isalnum()

    @op(name="string-lower")
    def lower(str: str) -> str:  # type: ignore[misc]
        return str.lower()

    @op(name="string-upper")
    def upper(str: str) -> str:  # type: ignore[misc]
        return str.upper()

    @op(name="string-slice")
    def slice(  # type: ignore[misc]
        str: str, begin: typing.Union[int, None], end: typing.Union[int, None] = None
    ) -> str:
        begin = 0 if begin is None else begin
        end = len(str) if end is None else end
        return str[begin:end]

    @op(name="string-replace")
    def replace(  # type: ignore[misc]
        str: str, sub: typing.Union[str, None], newSub: typing.Union[str, None]
    ) -> str:
        if sub is None:
            return str
        newSub = newSub or ""
        return str.replace(sub, newSub)

    @op(name="string-findAll")
    def findall(str: str, sub: typing.Union[str, None]) -> list[str]:  # type: ignore[misc]
        if sub is None:
            return []
        return [g[0] if isinstance(g, list) else g for g in re.findall(sub, str)]

    @op(name="string-strip")
    def strip(str: str) -> str:  # type: ignore[misc]
        return str.strip()

    @op(name="string-lStrip")
    def lstrip(str: str) -> str:  # type: ignore[misc]
        return str.lstrip()

    @op(name="string-rStrip")
    def rstrip(str: str) -> str:  # type: ignore[misc]
        return str.rstrip()

    @op()
    def format(
        self, named_items: dict[str, typing.Optional[typing.Union[str, int]]]
    ) -> str:
        return self.format(**named_items)

    @op(hidden=True)
    def json_parse_refine(self) -> types.Type:
        return types.TypeRegistry.type_of(_json_parse(self))  # type: ignore

    @op(refine_output_type=json_parse_refine)
    def json_parse(self) -> typing.Any:
        return _json_parse(self)  # type: ignore

    # I'd prefer json_parse(s, Type), or json_parse(s).cast(Type), but
    # for now this is easy. We know the output type is list at call time
    # so ops are dispatchable. And then we refine to get the full type.
    @op(refine_output_type=json_parse_refine)
    def json_parse_list(self) -> list[typing.Any]:
        res = _json_parse(self)  # type: ignore
        if not isinstance(res, list):
            raise ValueError("Expected a list")
        return res

    @op(name="string-toNumber", output_type=types.optional(types.Number()))
    def to_number(self):
        if self.isnumeric():
            return float(self)  # type: ignore
        return None


types.String.instance_class = String


@op(name="string-levenshtein", render_info={"type": "function"})
def levenshtein(str1: typing.Union[str, None], str2: typing.Union[str, None]) -> int:
    return _levenshtein(str1 or "", str2 or "")


def _levenshtein(str1: str, str2: str) -> int:
    """calculate the number of single-character edits between str1 and str2"""
    if len(str1) < len(str2):
        return _levenshtein(str2, str1)

    if len(str2) == 0:
        return len(str1)

    arr1 = np.array(tuple(str1))
    arr2 = np.array(tuple(str2))

    previous_row = np.arange(arr2.size + 1)
    for s in arr1:
        current_row = previous_row + 1
        current_row[1:] = np.minimum(
            current_row[1:],
            np.add(previous_row[:-1], arr2 != s),
        )
        current_row[1:] = np.minimum(
            current_row[1:],
            current_row[0:-1] + 1,
        )
        previous_row = current_row

    return previous_row[-1].item()



class Error(Exception): ...


class InternalError(Error): ...


class OpCallError(Error): ...


class WeaveTypeError(Error): ...


class WeaveSerializeError(Error): ...


class WeaveOpSerializeError(WeaveSerializeError): ...


class WeaveInitError(Error): ...


class WeaveDefinitionError(Error): ...


class WeaveWandbAuthenticationException(Error): ...



from urllib.parse import quote

from wandb import util as wb_util

from weave.legacy.weave import context_state, environment

BROWSE3_PATH = "browse3"
WEAVE_SLUG = "weave"


def remote_project_root_url(entity_name: str, project_name: str) -> str:
    return f"{wb_util.app_url(environment.wandb_base_url())}/{entity_name}/{quote(project_name)}"


def remote_project_weave_root_url(entity_name: str, project_name: str) -> str:
    return f"{remote_project_root_url(entity_name, project_name)}/{WEAVE_SLUG}"


def local_project_weave_root_url(entity_name: str, project_name: str) -> str:
    return f"http://localhost:3000/{BROWSE3_PATH}/{entity_name}/{quote(project_name)}"


def project_weave_root_url(entity_name: str, project_name: str) -> str:
    if context_state._use_local_urls.get():
        return local_project_weave_root_url(entity_name, project_name)
    else:
        return remote_project_weave_root_url(entity_name, project_name)


def op_version_path(
    entity_name: str, project_name: str, op_name: str, op_version: str
) -> str:
    return f"{project_weave_root_url(entity_name, project_name)}/ops/{op_name}/versions/{op_version}"


def object_version_path(
    entity_name: str, project_name: str, object_name: str, obj_version: str
) -> str:
    return f"{project_weave_root_url(entity_name, project_name)}/objects/{quote(object_name)}/versions/{obj_version}"


def call_path(entity_name: str, project_name: str, call_id: str) -> str:
    return f"{project_weave_root_url(entity_name, project_name)}/calls/{call_id}"


def redirect_call(entity_name: str, project_name: str, call_id: str) -> str:
    return f"{remote_project_root_url(entity_name, project_name)}/r/call/{call_id}"


def use_local_urls() -> None:
    context_state._use_local_urls.set(True)


def use_remote_urls() -> None:
    context_state._use_local_urls.set(False)



# Fuzz testing with the hypothesis library
#
# Quick summary: hypothesis is a fuzz testing framework. You
#   provide strategies for generating inputs, and assertions
#   for checking results. Hypothesis does a guided search
#   to look for bugs.
#
# Known Weave issues that we intentionally avoid here (but
# that will need to be fixed at some point):
# - Don't test large ints because we convert int64 to float64
#   in arrow sometimes, which truncates them.
# - We don't test float & int together in join2 because the Python
#   and arrow implementations differ in how they handle them.
#
# To debug and fix failing tests here:
# hypothesis will print output like this:
#   ValueError: ArrowWeaveList validation err: <UnknownType()>, null
#     Non-zero length array with UnknownType
#   Falsifying example: test_join2(
#       weave_no_cache=None,
#       list_lambda1=([0], lambda x: x),
#       list_lambda2=([], lambda x: x),
#       leftOuter=True,
#       rightOuter=False,
#   )
# This means that the test failed with the given inputs.
# You can force that specific example to run by adding
#   @example(
#       list_lambda1=([0], lambda x: x),
#       list_lambda2=([], lambda x: x),
#       leftOuter=True,
#       rightOuter=False,
#   )
# above the test function.

import dataclasses
import json
import os

import pyarrow as pa
import pytest
from hypothesis import example, given, settings
from hypothesis import strategies as st
from hypothesis.strategies import composite

import weave
from weave.legacy.weave import artifact_local, ops_arrow, ops_primitives, storage
from weave.legacy.weave.arrow import convert
from weave.legacy.weave.language_features.tagging import tag_store

# Jack this up to find more bugs.
EXAMPLES_PER_TEST = 100


@pytest.fixture(scope="module")
def weave_no_cache():
    old_cache = os.environ.get("WEAVE_NO_CACHE")
    os.environ["WEAVE_NO_CACHE"] = "1"
    yield
    if old_cache is None:
        del os.environ["WEAVE_NO_CACHE"]
    else:
        os.environ["WEAVE_NO_CACHE"] = old_cache


# MIN_INT64 = -(2**63)
# MAX_INT64 = 2**63 - 2  # why -2? If we use -1 we get an error

# We fail on union of large int and float, so don't use large ints for now
MIN_INT64 = -(2**32)
MAX_INT64 = 2**32


# make including float optional, because the behavior is different for
# arrow join v python join when both float and int and present. Rather
# than unifying, just disable the float case when doing join tests for now.
def obj_strategy(max_depth=3, include_float=True, include_none=True):
    basics = [
        st.integers(min_value=MIN_INT64, max_value=MAX_INT64),
        st.text(),
        st.booleans(),
    ]
    if include_float:
        basics.append(st.floats(allow_nan=False, allow_infinity=False))
    if include_none:
        basics.append(st.none())
    if max_depth <= 0:
        return st.one_of(*basics)
    else:
        return st.one_of(
            *basics,
            st.lists(
                obj_strategy(
                    max_depth - 1,
                    include_float=include_float,
                    include_none=include_none,
                ),
            ),
            st.dictionaries(
                st.sampled_from(["key", "foo", "bar"]),
                # Disable generating None values in dictionaries. Our arrow implementation
                # automatically adds Nones for missing keys in many cases. A known issue.
                obj_strategy(
                    max_depth - 1, include_float=include_float, include_none=False
                ),
                min_size=1,
                max_size=5,
            ),
        )


def list_strategy(max_depth=3, include_float=True):
    return st.lists(obj_strategy(max_depth, include_float))


def numeric_lambda_strategy():
    return st.sampled_from(
        [
            lambda x: x,
            lambda x: x * 2,
            lambda x: x - 1,
            lambda x: x + 1,
        ]
    )


@composite
def list_lambda_strategy(draw, include_float=True):
    l = draw(list_strategy(include_float=include_float))
    t = weave.type_of(l)
    ot = t.object_type
    if isinstance(ot, weave.types.TypedDict):
        return (l, lambda x: x["key"])
    elif isinstance(ot, weave.types.List):
        return (l, lambda x: x[0])
    elif isinstance(ot, weave.types.Number):
        return (l, draw(numeric_lambda_strategy()))
    else:
        return (l, lambda x: x)


def compare_py(val, orig):
    if isinstance(val, dict):
        assert isinstance(orig, dict)
        for k in orig:
            assert k in val
            compare_py(val[k], orig[k])
        for k in set(val) - set(orig):
            assert val[k] == None
    elif isinstance(val, list):
        assert isinstance(orig, list)
        assert len(val) == len(orig)
        for i in range(len(val)):
            compare_py(val[i], orig[i])
    else:
        assert val == orig


def fix_for_compare(x1):
    if isinstance(x1, list):
        return [fix_for_compare(x) for x in x1]
    elif isinstance(x1, dict):
        return {k: fix_for_compare(v) for k, v in x1.items()}
    elif dataclasses.is_dataclass(x1):
        return x1.__class__(
            **{k: fix_for_compare(v) for k, v in dataclasses.asdict(x1).items()}
        )
    elif isinstance(x1, artifact_local.LocalArtifact):
        return (x1.name, x1.version)
    return x1


@given(l1=st.lists(obj_strategy()))
@settings(max_examples=EXAMPLES_PER_TEST, deadline=None)
def test_to_from_arrow(l1):
    a1 = ops_arrow.to_arrow(l1)
    l1_result = a1.to_pylist_tagged()
    compare_py(l1_result, l1)


def rotate_list(l):
    if not l:
        return l
    return l[1:] + [l[0]]


def pa_equal_with_null(a1, a2):
    null_count = pa.compute.add(
        pa.compute.is_null(a1).cast(pa.int8()), pa.compute.is_null(a2).cast(pa.int8())
    )
    return pa.compute.if_else(
        pa.compute.equal(null_count, 2),
        pa.scalar(True),
        pa.compute.if_else(
            pa.compute.equal(null_count, 0), pa.compute.equal(a1, a2), pa.scalar(False)
        ),
    )


def compare_safe_equal(v1, v2):
    if isinstance(v1, dict):
        if not isinstance(v2, dict):
            return False
        for k in set(v1) | set(v2):
            if not compare_safe_equal(v1.get(k), v2.get(k)):
                return False
        return True
    elif isinstance(v1, list):
        if not isinstance(v2, list):
            return False
        if not len(v1) == len(v2):
            return False
        for i in range(len(v1)):
            if not compare_safe_equal(v1[i], v2[i]):
                return False
        return True
    elif isinstance(v1, bool):
        if not isinstance(v2, bool):
            return False
        return v1 == v2
    elif isinstance(v2, bool):
        if not isinstance(v1, bool):
            return False
        return v1 == v2
    elif isinstance(v1, (int, float)):
        if not isinstance(v2, (int, float)):
            return False
        return v1 == v2
    return v1 == v2


@given(l1=list_strategy())
@settings(max_examples=EXAMPLES_PER_TEST, deadline=None)
def test_to_compare_safe(l1):
    l2 = rotate_list(l1)
    a1 = ops_arrow.to_arrow(l1)
    a2 = ops_arrow.to_arrow(l2)
    safe1 = convert.to_compare_safe(a1)
    safe2 = convert.to_compare_safe(a2)
    safe1_arr = safe1._arrow_data
    safe2_arr = safe2._arrow_data
    # print("l1", l1)
    # print("l2", l2)
    # print("safe1_arr", safe1_arr)
    # print("safe2_arr", safe2_arr)
    if len(safe1_arr):
        # comparing to self is true!
        assert pa_equal_with_null(safe1_arr, safe1_arr).to_pylist() == [True] * len(
            safe1_arr
        )
    if len(safe2_arr):
        # comparing to self is true!
        assert pa_equal_with_null(safe2_arr, safe2_arr).to_pylist() == [True] * len(
            safe2_arr
        )
    if len(safe1_arr) and len(safe2_arr):
        assert pa_equal_with_null(safe1_arr, safe2_arr).to_pylist() == list(
            compare_safe_equal(i1, i2) for i1, i2 in zip(l1, l2)
        )


@given(l=st.lists(obj_strategy()))
@settings(max_examples=EXAMPLES_PER_TEST, deadline=None)
# @example(l=[None, {"foo": 0}, 0, {"foo": ""}])
@example(l=[None, {"foo": 15}, {"foo": "a"}])
def test_saveload(l, weave_no_cache):
    a = ops_arrow.to_arrow(l)
    a_ref = storage.save(a)
    a_again = storage.get(str(a_ref))
    assert a.object_type == a_again.object_type
    assert a._arrow_data == a_again._arrow_data


@given(l1=st.lists(obj_strategy()), l2=st.lists(obj_strategy()))
@settings(max_examples=EXAMPLES_PER_TEST, deadline=None)
def test_concat(l1, l2, weave_no_cache):
    a1 = ops_arrow.to_arrow(l1)
    l1_result = a1.to_pylist_tagged()
    compare_py(l1_result, l1)
    a1 = ops_arrow.to_arrow(l1)
    a2 = ops_arrow.to_arrow(l2)

    result = a1.concat(a2)
    expected = ops_arrow.to_arrow(l1 + l2)

    # assert result.object_type == expected.object_type
    assert fix_for_compare(result.to_pylist_tagged()) == fix_for_compare(
        expected.to_pylist_tagged()
    )


# @pytest.mark.flaky(reruns=3)
@given(
    list_lambda1=list_lambda_strategy(include_float=False),
    list_lambda2=list_lambda_strategy(include_float=False),
    leftOuter=st.booleans(),
    rightOuter=st.booleans(),
)
@settings(max_examples=EXAMPLES_PER_TEST, deadline=None)
def test_join2(list_lambda1, list_lambda2, leftOuter, rightOuter, weave_no_cache):
    l1, join1Fn = list_lambda1
    l2, join2Fn = list_lambda2
    # Call your join2_impl function
    alias1 = "alias1"
    alias2 = "alias2"
    l_result = weave.use(
        ops_primitives.join_2(
            l1, l2, join1Fn, join2Fn, alias1, alias2, leftOuter, rightOuter
        )
    )
    print("L1", l1)
    print("L2", l2)
    print("LEFT OUTER", leftOuter)
    print("RIGHT OUTER", rightOuter)
    a1 = ops_arrow.to_arrow(l1)
    a2 = ops_arrow.to_arrow(l2)
    arr_result = a1.join2(
        a2,
        join1Fn,
        join2Fn,
        alias1,
        alias2,
        leftOuter,
        rightOuter,
    )
    l_result_safe = ops_arrow.to_arrow(l_result).to_pylist_tagged()
    sorted_l_result_safe = sorted(l_result_safe, key=lambda x: json.dumps(x))
    sorted_arr_result = sorted(
        arr_result.to_pylist_tagged(), key=lambda x: json.dumps(x)
    )
    # print("L_RESULT_ORIG", l_result)
    # print("L_RESULT", sorted_l_result_safe)
    # print("ARR_RESULT", sorted_arr_result)
    with tag_store.new_tagging_context():
        compare_py(sorted_arr_result, sorted_l_result_safe)

    # # Add your assertions here
    # # For example, you can test the length of the result array


if __name__ == "__main__":
    test_to_compare_safe()



import importlib
import typing

import weave
from weave.trace.op_extensions.accumulator import add_accumulator
from weave.trace.patcher import MultiPatcher, SymbolPatcher

if typing.TYPE_CHECKING:
    from mistralai.models import (
        ChatCompletionResponse,
        CompletionEvent,
    )


def mistral_accumulator(
    acc: typing.Optional["ChatCompletionResponse"],
    value: "CompletionEvent",
) -> "ChatCompletionResponse":
    # This import should be safe at this point
    from mistralai.models import (
        AssistantMessage,
        ChatCompletionChoice,
        ChatCompletionResponse,
        UsageInfo,
    )

    value = value.data
    if acc is None:
        acc = ChatCompletionResponse(
            id=value.id,
            object=value.object,
            created=value.created,
            model=value.model,
            choices=[],
            usage=UsageInfo(prompt_tokens=0, total_tokens=0, completion_tokens=0),
        )

    # Merge in the usage info
    if value.usage is not None:
        acc.usage.prompt_tokens += value.usage.prompt_tokens
        acc.usage.total_tokens += value.usage.total_tokens
        if acc.usage.completion_tokens is None:
            acc.usage.completion_tokens = value.usage.completion_tokens
        else:
            acc.usage.completion_tokens += value.usage.completion_tokens

    # Loop through the choices and add their deltas
    for delta_choice in value.choices:
        while delta_choice.index >= len(acc.choices):
            acc.choices.append(
                ChatCompletionChoice(
                    index=len(acc.choices),
                    message=AssistantMessage(content=""),
                    finish_reason="stop",
                )
            )

        if acc.choices is None:
            return acc

        target_choice: ChatCompletionChoice = acc.choices[delta_choice.index]

        if target_choice is None:
            return acc

        target_choice.finish_reason = (
            delta_choice.finish_reason or target_choice.finish_reason
        )

        target_choice.message.role = (
            delta_choice.delta.role or target_choice.message.role
        )
        target_choice.message.content += delta_choice.delta.content or ""
        if delta_choice.delta.tool_calls:
            if target_choice.message.tool_calls is None:
                target_choice.message.tool_calls = []
            target_choice.message.tool_calls += delta_choice.delta.tool_calls

    return acc


def mistral_stream_wrapper(name: str) -> typing.Callable:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        op = weave.op()(fn)
        acc_op = add_accumulator(op, lambda inputs: mistral_accumulator)  # type: ignore
        acc_op.name = name  # type: ignore
        return acc_op

    return wrapper


def mistral_wrapper(name: str) -> typing.Callable:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        op = weave.op()(fn)
        op.name = name  # type: ignore
        return op

    return wrapper


mistral_patcher = MultiPatcher(
    [
        # Patch the sync, non-streaming chat method
        SymbolPatcher(
            lambda: importlib.import_module("mistralai.chat"),
            "Chat.complete",
            mistral_wrapper(name="Mistral.chat.complete"),
        ),
        # Patch the sync, streaming chat method
        SymbolPatcher(
            lambda: importlib.import_module("mistralai.chat"),
            "Chat.stream",
            mistral_stream_wrapper(name="Mistral.chat.stream"),
        ),
        # Patch the async, non-streaming chat method
        SymbolPatcher(
            lambda: importlib.import_module("mistralai.chat"),
            "Chat.complete_async",
            mistral_wrapper(name="Mistral.chat.complete_async"),
        ),
        # Patch the async, streaming chat method
        SymbolPatcher(
            lambda: importlib.import_module("mistralai.chat"),
            "Chat.stream_async",
            mistral_stream_wrapper(name="Mistral.chat.stream_async"),
        ),
    ]
)



import pickle

import spacy as spacy_lib

import weave


class SpacyDocType(weave.types.Type):
    instance_classes = spacy_lib.tokens.doc.Doc

    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.pickle", binary=True) as f:
            pickle.dump(obj, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.pickle", binary=True) as f:
            return pickle.load(f)


@weave.op(render_info={"type": "function"})
def spacy(text: str) -> spacy_lib.tokens.doc.Doc:
    # TODO: Make this into a package that loads all the models from spacy,
    # has types, and supports different Components (similar to HF). For now,
    # this is just a simple english model
    import spacy as spacy_lib

    nlp = spacy_lib.load("en_core_web_sm")
    return nlp(text)


@weave.op()
def spacy_doc_dep_to_html(
    spacy_doc: spacy_lib.tokens.doc.Doc,
) -> weave.legacy.weave.ops.Html:
    from spacy import displacy

    html = displacy.render(
        list(spacy_doc.sents), style="dep", jupyter=False, options={"compact": True}
    )
    return weave.legacy.weave.ops.Html(html)


@weave.op()
def spacy_doc_ent_to_html(
    spacy_doc: spacy_lib.tokens.doc.Doc,
) -> weave.legacy.weave.ops.Html:
    from spacy import displacy

    html = displacy.render(spacy_doc, style="ent", jupyter=False)
    return weave.legacy.weave.ops.Html(html)


@weave.type()
class SpacyDocPanel(weave.Panel):
    id = "SpacyDocPanel"
    input_node: weave.Node[spacy_lib.tokens.doc.Doc]

    @weave.op()
    def render(self) -> weave.legacy.weave.panels.Card:
        return weave.legacy.weave.panels.Card(
            title="Spacy Visualization",
            subtitle="",
            content=[
                weave.legacy.weave.panels.CardTab(
                    name="Dependencies",
                    content=weave.legacy.weave.panels.PanelHtml(spacy_doc_dep_to_html(self.input_node)),  # type: ignore
                ),
                weave.legacy.weave.panels.CardTab(
                    name="Named Entities",
                    content=weave.legacy.weave.panels.PanelHtml(spacy_doc_ent_to_html(self.input_node)),  # type: ignore
                ),
            ],
        )



import dataclasses
import typing

import torch
from torch import nn, optim
from torch.utils.data import Dataset
from torchvision import transforms

import weave
from weave.legacy.weave import context_state as _context
from weave.legacy.weave.ecosystem import pytorch

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")


@dataclasses.dataclass(frozen=True)
class ModelType(weave.types.ObjectType):
    input_type: weave.types.Type = weave.types.Any()
    output_type: weave.types.Type = weave.types.Any()

    def property_types(self):
        # Huh... this has the same interface as any function...
        return {
            "input_type": self.input_type,
            "output_type": self.output_type,
            "pred_fn": pytorch.TorchModelType(),
        }


@weave.weave_class(weave_type=ModelType)
@dataclasses.dataclass()
class Model:
    input_type: weave.types.Type
    output_type: weave.types.Type

    pred_fn: torch.nn.Sequential  # just the torch model above for now

    @weave.op(
        input_type={
            "self": ModelType(),
            "X": lambda input_type: input_type["self"].input_type,
        },
        # TODO
        # output_type=lambda input_type: weave.types.List(
        #     weave.types.TypedDict(
        #         {
        #             "X": input_type["self"].input_type.object_type,
        #             "y": input_type["self"].output_type.object_type,
        #         }
        #     )
        # ),
        # I hardcoded the output_type of predict and train just to show the example
        # working.
        output_type=weave.types.List(
            weave.types.TypedDict(
                {
                    "X": weave.legacy.weave.ops.image.PILImageType(),  # type: ignore
                    "y": weave.types.Int(),
                }
            )
        ),
    )
    def predict(self, X):
        # Note, this is a copy of the transform in TorchMnistDataset!
        transform = transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
        )
        inputs = [transform(x) for x in X]
        class_preds = self.pred_fn(torch.stack(inputs))
        _, final_preds = torch.max(class_preds.data, 1)
        final_preds = final_preds.tolist()
        # TODO: do a faster column-oriented version
        rows = []
        for x, y in zip(X, final_preds):
            rows.append({"X": x, "y": y})
        return rows


ModelType.instance_class = Model
ModelType.instance_classes = Model


class TorchMnistTrainConfig(typing.TypedDict):
    fc_layer_size: int
    dropout: float
    epochs: int
    learning_rate: float
    batch_size: int


class TorchMnistDataset(Dataset):
    transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
    )

    def __init__(self, X, y):
        self.X = X
        self.y = y

    def __len__(self):
        return len(self.X)

    def __getitem__(self, i):
        return self.transform(self.X[i]), self.y[i]


def build_network(fc_layer_size, dropout):
    network = nn.Sequential(  # fully-connected, single hidden layer
        nn.Flatten(),
        nn.Linear(784, fc_layer_size),
        nn.ReLU(),
        nn.Dropout(dropout),
        nn.Linear(fc_layer_size, 10),
        nn.LogSoftmax(dim=1),
    )
    return network.to(DEVICE)


def train_epoch(network, loader, optimizer):
    cumu_loss = 0
    for _, (data, target) in enumerate(loader):
        data, target = data.to(DEVICE), target.to(DEVICE)
        optimizer.zero_grad()
        loss = nn.functional.nll_loss(network(data), target)
        cumu_loss += loss.item()
        loss.backward()
        optimizer.step()
    return cumu_loss / len(loader)


@weave.op(
    render_info={"type": "function"},
    input_type={
        "X": weave.types.List(weave.legacy.weave.ops.image.PILImageType()),  # type: ignore
        "y": weave.types.List(weave.types.Int()),  # TODO: class enum?
    },
    # TODO: WeaveJS doesn't support callable output type yet.
    # output_type=lambda input_type: ModelType(input_type["X"], input_type["y"]),
    output_type=ModelType(),
)
def train(X, y, config: TorchMnistTrainConfig):
    loader = torch.utils.data.DataLoader(
        TorchMnistDataset(X, y), batch_size=config["batch_size"]
    )
    network = build_network(config["fc_layer_size"], config["dropout"])
    optimizer = optim.Adam(network.parameters(), lr=config["learning_rate"])
    for epoch in range(config["epochs"]):
        print("Epoch: %s" % epoch)
        avg_loss = train_epoch(network, loader, optimizer)
    return Model(weave.type_of(X), weave.type_of(y), network)



from weave import types
from weave.legacy.weave import graph

from . import debug_types


def assign_type_weave0_weave1(w0_type: types.Type, w1_type: types.Type) -> bool:  # type: ignore[empty-body]
    pass


def check_weave0_compile_result(
    weave0_nodes: list[graph.Node], weave1_nodes: list[graph.Node]
) -> None:
    weave0_flat: list[graph.OutputNode] = [
        n for n in graph.all_nodes_full(weave0_nodes) if isinstance(n, graph.OutputNode)
    ]
    weave1_flat: list[graph.OutputNode] = [
        n for n in graph.all_nodes_full(weave1_nodes) if isinstance(n, graph.OutputNode)
    ]
    if len(weave0_flat) != len(weave1_flat):
        print("weave0 and weave1 have different lengths")
        raise ValueError
    for op0, op1 in zip(weave0_flat, weave1_flat):
        if op1.from_op.name != "gqlroot-wbgqlquery":
            if op0.from_op.friendly_name != op1.from_op.friendly_name:
                print("weave0 and weave1 have different op names", op0, op1)
                raise ValueError
            if not op0.type.assign_type(op1.type):
                print(
                    "weave1 type not assignable to weave0 type for op0",
                    op0.from_op.name,
                    op0.type,
                    op1.type,
                )
                print(debug_types.why_not_assignable(op0.type, op1.type))
                raise ValueError



import pytest

import weave
from weave.legacy.weave import api, context_state, environment, errors, ops


def test_dir():
    testdir = ops.local_path("testdata/test_dir")
    # TODO: causes test to fail because of serialization mutatio

    size = api.use(testdir.size)
    assert size == 111
    dir = api.use(testdir)
    assert len(dir.dirs) == 1
    assert len(dir.dirs["sub_dir"].dirs) == 0
    assert len(dir.dirs["sub_dir"].files) == 1
    assert len(dir.files) == 2


def test_nonexistent_file():
    with pytest.raises(FileNotFoundError):
        ops.local_path("fake_file_.idontexist")



from weave.legacy.weave import gql_to_weave
from weave.legacy.weave import weave_types as types


def test_simple_query():
    query = """
    query {
        project(name: "test") {
            id
            name
        }
    }
    """
    return_type = gql_to_weave.get_query_weave_type(query)
    assert return_type == types.TypedDict(
        {
            "project": types.TypedDict(
                {
                    "id": types.String(),
                    "name": types.String(),
                }
            )
        }
    )


def test_artifact_query():
    query = """
        query WeavePythonCG {
        project_8d1592567720841659de23c02c97d594: project(
            name: "p_0"
            entityName: "e_0"
        ) {
            id
            name
            runs_261949318143369aa6c158af92afee03: runs(
            first: 100
            filters: "{}"
            order: "-createdAt"
            ) {
            edges {
                node {
                id
                name
                summaryMetricsSubset: summaryMetrics(
                    keys: ["Tables/NMS_0.45_IOU_0.5"]
                )
                project {
                    id
                    name
                    entity {
                    id
                    name
                    }
                }
                }
            }
            }
        }
        }
        """
    return_type = gql_to_weave.get_query_weave_type(query)
    assert return_type == types.TypedDict(
        {
            "project_8d1592567720841659de23c02c97d594": types.TypedDict(
                {
                    "id": types.String(),
                    "name": types.String(),
                    "runs_261949318143369aa6c158af92afee03": types.TypedDict(
                        {
                            "edges": types.List(
                                types.TypedDict(
                                    {
                                        "node": types.TypedDict(
                                            {
                                                "id": types.String(),
                                                "name": types.String(),
                                                "summaryMetricsSubset": types.optional(
                                                    types.String()
                                                ),
                                                "project": types.TypedDict(
                                                    {
                                                        "id": types.String(),
                                                        "name": types.String(),
                                                        "entity": types.TypedDict(
                                                            {
                                                                "id": types.String(),
                                                                "name": types.String(),
                                                            }
                                                        ),
                                                    }
                                                ),
                                            }
                                        )
                                    }
                                )
                            )
                        }
                    ),
                }
            )
        }
    )


def test_multi_root_query():
    query = """
        query WeavePythonCG{
            project_8d1592567720841659de23c02c97d594:project(name:"p_0" entityName:"e_0"){
                id
                name
                createdAt
            }
            project_3c237e5b25fed9a705b21513dd7921c6:project(name:"p_1" entityName:"e_1"){
                id
                name
                runs_c1233b7003317090ab5e2a75db4ad965:runs(first:100){
                    edges{
                        node{
                            id
                            name
                        }
                    }
                }
            }
            instance{
                projects_500:projects(limit:500){
                    edges{
                        node{
                            id
                            name
                            createdAt
                        }
                    }
                }
            }
        }
        """

    return_type = gql_to_weave.get_query_weave_type(query)
    assert return_type == types.TypedDict(
        {
            "project_8d1592567720841659de23c02c97d594": types.TypedDict(
                {
                    "id": types.String(),
                    "name": types.String(),
                    "createdAt": types.Timestamp(),
                }
            ),
            "project_3c237e5b25fed9a705b21513dd7921c6": types.TypedDict(
                {
                    "id": types.String(),
                    "name": types.String(),
                    "runs_c1233b7003317090ab5e2a75db4ad965": types.TypedDict(
                        {
                            "edges": types.List(
                                types.TypedDict(
                                    {
                                        "node": types.TypedDict(
                                            {
                                                "id": types.String(),
                                                "name": types.String(),
                                            }
                                        )
                                    }
                                )
                            )
                        }
                    ),
                }
            ),
            "instance": types.TypedDict(
                {
                    "projects_500": types.TypedDict(
                        {
                            "edges": types.List(
                                types.TypedDict(
                                    {
                                        "node": types.TypedDict(
                                            {
                                                "id": types.String(),
                                                "name": types.String(),
                                                "createdAt": types.Timestamp(),
                                            }
                                        )
                                    }
                                )
                            )
                        }
                    )
                }
            ),
        }
    )


def test_inline_fragments():
    query = """query WeavePythonCG {
  project_518fa79465d8ffaeb91015dce87e092f: project(
    name: "mendeleev"
    entityName: "stacey"
  ) {
    id
    name
    artifactType_46d22fef09db004187bb8da4b5e98c58: artifactType(
      name: "test_results"
    ) {
      id
      name
      artifactCollections_c1233b7003317090ab5e2a75db4ad965: artifactCollections(
        first: 100
      ) {
        edges {
          node {
            id
            name

            artifacts_c1233b7003317090ab5e2a75db4ad965: artifacts(first: 100) {
              edges {
                node {
                  id
                  createdBy {
                    __typename
                    ... on Run {
                      id
                      name
                      displayName
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
}
    """

    return_type = gql_to_weave.get_query_weave_type(query)
    expected = types.TypedDict(
        {
            "project_518fa79465d8ffaeb91015dce87e092f": types.TypedDict(
                {
                    "id": types.String(),
                    "name": types.String(),
                    "artifactType_46d22fef09db004187bb8da4b5e98c58": types.TypedDict(
                        {
                            "id": types.String(),
                            "name": types.String(),
                            "artifactCollections_c1233b7003317090ab5e2a75db4ad965": types.TypedDict(
                                {
                                    "edges": types.List(
                                        types.TypedDict(
                                            {
                                                "node": types.TypedDict(
                                                    {
                                                        "id": types.String(),
                                                        "name": types.String(),
                                                        "artifacts_c1233b7003317090ab5e2a75db4ad965": types.TypedDict(
                                                            {
                                                                "edges": types.List(
                                                                    types.TypedDict(
                                                                        {
                                                                            "node": types.TypedDict(
                                                                                {
                                                                                    "id": types.String(),
                                                                                    "createdBy": types.UnionType(
                                                                                        types.TypedDict(
                                                                                            {
                                                                                                "__typename": types.Const(
                                                                                                    types.String(),
                                                                                                    "Run",
                                                                                                ),
                                                                                                "id": types.String(),
                                                                                                "name": types.String(),
                                                                                                "displayName": types.optional(
                                                                                                    types.String()
                                                                                                ),
                                                                                            }
                                                                                        ),
                                                                                        types.TypedDict(
                                                                                            {
                                                                                                "__typename": types.Const(
                                                                                                    types.String(),
                                                                                                    "User",
                                                                                                ),
                                                                                            }
                                                                                        ),
                                                                                    ),
                                                                                }
                                                                            )
                                                                        }
                                                                    )
                                                                )
                                                            }
                                                        ),
                                                    }
                                                )
                                            }
                                        )
                                    )
                                }
                            ),
                        }
                    ),
                }
            )
        }
    )

    assert return_type == expected


def test_json_array():
    query = """query WeavePythonCG {
  project_23045de0d7920e9b08772d538b4b9da9: project(
    name: "oai-mon-syn-1"
    entityName: "shawn"
  ) {
    id
    name
    run_1dbf019d4314adc6f3bc206824769929: run(name: "table100-2") {
      id
      name
      historyKeys
      sampledParquetHistory: parquetHistory(
        liveKeys: [
          "output.id"
          "output.object"
          "timestamp"
          "output.model"
          "output.choices"
          "_step"
          "summary.prompt_tokens"
          "output.created"
          "summary.completion_tokens"
          "summary.latency_s"
          "summary.total_tokens"
        ]
      ) {
        liveData
        parquetUrls
      }
      project {
        id
        name
        entity {
          id
          name
        }
      }
    }
  }
}"""
    return_type = gql_to_weave.get_query_weave_type(query)
    expected = types.TypedDict(
        {
            "project_23045de0d7920e9b08772d538b4b9da9": types.TypedDict(
                property_types={
                    "id": types.String(),
                    "name": types.String(),
                    "run_1dbf019d4314adc6f3bc206824769929": types.TypedDict(
                        property_types={
                            "id": types.String(),
                            "name": types.String(),
                            "historyKeys": types.optional(types.String()),
                            "sampledParquetHistory": types.TypedDict(
                                property_types={
                                    # important that liveData is a string
                                    "liveData": types.String(),
                                    "parquetUrls": types.List(
                                        object_type=types.String()
                                    ),
                                },
                                not_required_keys=set(),
                            ),
                            "project": types.TypedDict(
                                property_types={
                                    "id": types.String(),
                                    "name": types.String(),
                                    "entity": types.TypedDict(
                                        property_types={
                                            "id": types.String(),
                                            "name": types.String(),
                                        },
                                        not_required_keys=set(),
                                    ),
                                },
                                not_required_keys=set(),
                            ),
                        },
                        not_required_keys=set(),
                    ),
                },
                not_required_keys=set(),
            )
        },
        not_required_keys=set(),
    )

    assert return_type == expected



"""The top-level functions for Weave Query API."""

import typing

from weave.legacy.weave import graph as _graph
from weave.legacy.weave.graph import Node

# If this is not imported, serialization of Weave Nodes is incorrect!
from weave.legacy.weave import graph_mapper as _graph_mapper

from . import storage as _storage
from . import ref_base as _ref_base
from weave.legacy.weave import wandb_api as _wandb_api

from weave.legacy.weave import weave_internal as _weave_internal

from weave.legacy.weave import util as _util

from weave.legacy.weave import context as _context
from ...trace import weave_init as _weave_init
from ...trace import weave_client as _weave_client

# exposed as part of api
from weave.legacy.weave import weave_types as types

# needed to enable automatic numpy serialization
from . import types_numpy as _types_numpy

from weave.legacy.weave import errors
from weave.legacy.weave.decorators import weave_class, mutation, type

from weave.legacy.weave import usage_analytics
from weave.legacy.weave.context import (
    use_fixed_server_port,
    use_frontend_devmode,
    # eager_execution,
    use_lazy_execution,
)

from weave.legacy.weave.panel import Panel

from weave.legacy.weave.arrow.list_ import ArrowWeaveList as WeaveList

# TODO: This is here because the op overloaded...
from weave.trace.op import op  # noqa: F401

def save(node_or_obj, name=None):  # type: ignore
    from weave.legacy.weave.ops_primitives.weave_api import get, save

    if isinstance(node_or_obj, _graph.Node):
        return save(node_or_obj, name=name)
    else:
        # If the user does not provide a branch, then we explicitly set it to
        # the default branch, "latest".
        branch = None
        name_contains_branch = name is not None and ":" in name
        if not name_contains_branch:
            branch = "latest"
        ref = _storage.save(node_or_obj, name=name, branch=branch)
        if name is None:
            # if the user didn't provide a name, the returned reference
            # will be to the specific version
            uri = ref.uri
        else:
            # otherwise the reference will be to whatever branch was provided
            # or the "latest" branch if only a name was provided.
            uri = ref.branch_uri
        return get(str(uri))


def get(ref_str):  # type: ignore
    obj = _storage.get(ref_str)
    ref = typing.cast(_ref_base.Ref, _storage._get_ref(obj))
    return _weave_internal.make_const_node(ref.type, obj)


def use(nodes, client=None):  # type: ignore
    usage_analytics.use_called()
    if client is None:
        client = _context.get_client()
    return _weave_internal.use(nodes, client)


def _get_ref(obj):  # type: ignore
    if isinstance(obj, _storage.Ref):
        return obj
    ref = _storage.get_ref(obj)
    if ref is None:
        raise errors.WeaveApiError("obj is not a weave object: %s" % obj)
    return ref


def versions(obj):  # type: ignore
    if isinstance(obj, _graph.ConstNode):
        obj = obj.val
    elif isinstance(obj, _graph.OutputNode):
        obj = use(obj)
    ref = _get_ref(obj)
    return ref.versions()  # type: ignore


def expr(obj):  # type: ignore
    ref = _get_ref(obj)
    return _trace.get_obj_expr(ref)


def type_of(obj: typing.Any) -> types.Type:  # type: ignore
    return types.TypeRegistry.type_of(obj)


# def weave(obj: typing.Any) -> RuntimeConstNode:
#     return _weave_internal.make_const_node(type_of(obj), obj)  # type: ignore


def from_pandas(df):  # type: ignore
    return _ops.pandas_to_awl(df)


__all__ = [
    # These seem to be important imports for query service...
    # TODO: Remove as many as possible...
    "_graph",
    "Node",
    "_graph_mapper",
    "_storage",
    "_ref_base",
    "_wandb_api",
    "_weave_internal",
    "_util",
    "_context",
    "_weave_init",
    "_weave_client",
    "types",
    "_types_numpy",
    "errors",
    "mutation",
    "weave_class",
    "type",
    "usage_analytics",
    "use_fixed_server_port",
    "use_frontend_devmode",
    "use_lazy_execution",
    "Panel",
    "WeaveList",
    # These are the actual functions declared
    "save",
    "get",
    "use",
    "_get_ref",
    "versions",
    "expr",
    "type_of",
    "from_pandas",
]



import typing

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import mappers
from weave.legacy.weave.partial_object import PartialObjectType


class TypedDictMapper(mappers.Mapper):
    def __init__(self, type_: types.TypedDict, mapper, artifact, path=[]):
        self.type = type_
        self._artifact = artifact
        self.path = path
        prop_serializers = {}
        for property_key, property_type in type_.property_types.items():
            prop_serializer = mapper(
                property_type, artifact, path=path + [property_key]
            )
            prop_serializers[property_key] = prop_serializer
        self._property_serializers = prop_serializers


class GQLMapper(mappers.Mapper):
    def __init__(self, type_: types.Type, mapper, artifact, path):
        self.type = typing.cast(PartialObjectType, type_)
        self._artifact = artifact
        self.path = path
        self.mapper = mapper
        prop_serializers = {}
        for property_key, property_type in self.type.keys.items():
            prop_serializer = mapper(
                property_type, artifact, path=path + [property_key]
            )
            prop_serializers[property_key] = prop_serializer
        self._property_serializers: dict[str, mappers.Mapper] = prop_serializers


class DictMapper(mappers.Mapper):
    def __init__(self, type_: types.Dict, mapper, artifact, path=[]):
        self.type = type_
        self.key_serializer = mapper(type_.key_type, artifact, path)
        self.value_serializer = mapper(type_.object_type, artifact, path)


class ObjectMapper(mappers.Mapper):
    def __init__(self, type_, mapper, artifact, path=[]):
        self.type = type_
        self._artifact = artifact
        prop_serializers = {}
        for property_key, property_type in type_.property_types().items():
            prop_serializer = mapper(
                property_type, artifact, path=path + [property_key]
            )
            prop_serializers[property_key] = prop_serializer
        self._obj_type = type_
        self._property_serializers = prop_serializers


class ListMapper(mappers.Mapper):
    def __init__(self, type_, mapper, artifact, path=[]):
        self.type = type_
        self._object_type = mapper(type_.object_type, artifact, path=path)


class UnionMapper(mappers.Mapper):
    # If true, this is a mapper for an optional type,
    # like <int> | <none> or <str> | <none>
    is_single_object_nullable: bool

    def __init__(self, type_, mapper, artifact, path=[]):
        self.type = type_
        self._member_mappers = [
            mapper(mem_type, artifact, path=path) for mem_type in type_.members
        ]

        self.is_single_object_nullable = False
        for mapper in self._member_mappers:
            if (
                isinstance(mapper.type, types.NoneType)
                and len(self._member_mappers) == 2
            ):
                self.is_single_object_nullable = True


class RefMapper(mappers.Mapper):
    def __init__(self, type_, mapper, artifact, path=[]):
        self.type = type_


class ConstMapper(mappers.Mapper):
    def __init__(self, type_, mapper, artifact, path):
        self.type = type_
        self._val_mapper = mapper(type_.val_type, artifact, path=path)



# history2 is depcreated, don't use it. It's only used in RunChain,
# which is just in demo form now. When we go back to working on RunChain
# we'll remove history2

import json

import pyarrow as pa

from weave.legacy.weave import engine_trace
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op, use
from weave.legacy.weave import artifact_mem, gql_json_cache
from weave.legacy.weave.arrow import convert
from weave.legacy.weave.arrow.list_ import ArrowWeaveList, ArrowWeaveListType
from weave.legacy.weave.gql_op_plugin import wb_gql_op_plugin
from weave.legacy.weave.ops_arrow.list_ops import concat
from weave.legacy.weave.ops_domain import wb_domain_types as wdt
from weave.legacy.weave.ops_domain import wb_util, wbmedia
from weave.legacy.weave.ops_domain.run_history import history_op_common

tracer = engine_trace.tracer()


@op(
    render_info={"type": "function"},
    plugins=wb_gql_op_plugin(lambda inputs, inner: "historyKeys"),
    hidden=True,
)
def refine_history2_type(run: wdt.Run) -> types.Type:
    return ArrowWeaveListType(history_op_common.refine_history_type(run))


@op(
    render_info={"type": "function"},
    plugins=wb_gql_op_plugin(lambda inputs, inner: "historyKeys"),
    hidden=True,
)
def refine_history2_with_columns_type(
    run: wdt.Run, history_cols: list[str]
) -> types.Type:
    return ArrowWeaveListType(
        history_op_common.refine_history_type(
            run, columns=history_op_common.get_full_columns_prefixed(run, history_cols)
        )
    )


@op(
    name="run-history2_with_columns",
    refine_output_type=refine_history2_with_columns_type,
    plugins=wb_gql_op_plugin(history_op_common.make_run_history_gql_field),
    output_type=ArrowWeaveListType(types.TypedDict({})),
    hidden=True,
)
def history2_with_columns(run: wdt.Run, history_cols: list[str]):
    return _get_history2(
        run, history_op_common.get_full_columns_prefixed(run, history_cols)
    )


# DEPRECATED: see comment at top.
# There is a bug in history2 where we mutate gql results, which is not allowed.
# now that we've removed deepcopy from gql_json_cache.use_json. But that's ok
# because this is deprecated and only used in a demo notebook.
@op(
    name="run-history2",
    refine_output_type=refine_history2_type,
    plugins=wb_gql_op_plugin(history_op_common.make_run_history_gql_field),
    output_type=ArrowWeaveListType(types.TypedDict({})),
    hidden=True,
)
def history2(run: wdt.Run):
    # We return mock data here since we will be replaced with the `_with_columns`
    # version in a compile pass if specific columns are needed
    return history_op_common.mock_history_rows(run)


def _get_history2(run: wdt.Run, columns=None):
    with tracer.trace("get_history") as span:
        span.set_tag("history_version", 2, 2)
    """Dont read binary columns. Keep everything in arrow. Faster, but not as full featured as get_history"""
    scalar_keys = history_op_common.history_keys(run)
    columns = [c for c in columns if c in scalar_keys]
    parquet_history = history_op_common.read_history_parquet(run, columns=columns)

    _object_type = history_op_common.refine_history_type(run, columns=columns)

    _history_type = types.List(_object_type)

    run_path = wb_util.RunPath(
        run["project"]["entity"]["name"],
        run["project"]["name"],
        run["name"],
    )

    # turn the liveset into an arrow table. the liveset is a list of json objects
    # we read from the cache via use_json()
    live_data = gql_json_cache.use_json(run["sampledParquetHistory"]["liveData"])

    # This is a nono! Mutating the gql results. But that's ok because this is deprecated,
    # see module comment at top.
    for row in live_data:
        for colname in columns:
            if colname not in row:
                row[colname] = None
            else:
                row[colname] = wb_util._process_run_dict_item(row[colname], run_path)

    artifact = artifact_mem.MemArtifact()
    # turn live data into arrow
    if live_data is not None and len(live_data) > 0:
        with tracer.trace("live_data_to_arrow"):
            live_data = convert.to_arrow(live_data, _history_type, artifact=artifact)
    else:
        live_data = []

    # get binary fields from history schema - these are serialized json
    if parquet_history is not None:
        fields = [field.name for field in parquet_history.schema]
        binary_fields = {
            field.name
            for field in parquet_history.schema
            if pa.types.is_binary(field.type)
        }

        # deserialize json if any is present
        with tracer.trace("process_non_basic_fields"):
            for field in columns:
                if field in binary_fields or not (
                    types.optional(types.BasicType()).assign_type(
                        _object_type.property_types[field]  # type: ignore
                    )
                    or wbmedia.ImageArtifactFileRefType().assign_type(
                        _object_type.property_types[field]  # type: ignore
                    )
                ):
                    pq_col = parquet_history[field].to_pylist()
                    for i, item in enumerate(pq_col):
                        if item is not None:
                            pq_col[i] = wb_util._process_run_dict_item(
                                json.loads(item) if field in binary_fields else item,
                                run_path,
                            )

                    awl = convert.to_arrow(
                        pq_col,
                        types.List(_object_type.property_types[field]),
                        artifact=artifact,
                    )
                    new_col = pa.chunked_array([awl._arrow_data])
                    parquet_history = parquet_history.set_column(
                        fields.index(field), field, new_col
                    )

    if parquet_history is not None and len(parquet_history) > 0:
        with tracer.trace("parquet_history_to_arrow"):
            parquet_history = ArrowWeaveList(
                parquet_history,
                _object_type,
                artifact=artifact,
            )
    else:
        parquet_history = []

    if len(live_data) == 0 and len(parquet_history) == 0:
        return ArrowWeaveList(pa.array([]), _object_type, artifact=artifact)
    elif len(live_data) == 0:
        return parquet_history
    elif len(parquet_history) == 0:
        return live_data
    return use(concat([parquet_history, live_data]))



import json
import typing
from dataclasses import dataclass, field

from . import stitch


@dataclass
class InputProvider:
    raw: dict[str, typing.Any]
    _dumps_cache: dict[str, str] = field(init=False, repr=False, default_factory=dict)

    def __getitem__(self, key: str) -> typing.Any:
        if key not in self.raw:
            raise KeyError(f"Input {key} not found")
        if key not in self._dumps_cache:
            self._dumps_cache[key] = json.dumps(
                self.raw[key] if self.raw[key] != None else ""
            )
        return self._dumps_cache[key]


@dataclass
class InputAndStitchProvider(InputProvider):
    stitched_obj: stitch.ObjectRecorder



"""Defines the Op protocol and related functions."""

import inspect
import logging
import sys
import typing
from functools import partial, wraps
from types import MethodType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Dict,
    Mapping,
    Optional,
    Protocol,
    Union,
    cast,
    overload,
    runtime_checkable,
)

from weave.legacy.weave import context_state
from weave.trace import box, call_context, settings
from weave.trace.client_context import weave_client as weave_client_context
from weave.trace.context import call_attributes, get_raise_on_captured_errors
from weave.trace.errors import OpCallError
from weave.trace.op_extensions.log_once import log_once
from weave.trace.refs import ObjectRef

logger = logging.getLogger(__name__)

from .constants import TRACE_CALL_EMOJI

if TYPE_CHECKING:
    from weave.trace.weave_client import Call, CallsIter

try:
    from openai._types import NOT_GIVEN as OPENAI_NOT_GIVEN
except ImportError:
    OPENAI_NOT_GIVEN = None

try:
    from cohere.base_client import COHERE_NOT_GIVEN
except ImportError:
    COHERE_NOT_GIVEN = None

try:
    from anthropic._types import NOT_GIVEN as ANTHROPIC_NOT_GIVEN
except ImportError:
    ANTHROPIC_NOT_GIVEN = None

try:
    # https://github.com/search?q=repo:mistralai/client-python%20Final&type=code
    from mistralai.types.basemodel import UNSET  # type: ignore

    MISTRAL_NOT_GIVEN = UNSET  # type: ignore
except ImportError:
    MISTRAL_NOT_GIVEN = None

MISTRAL_NOT_GIVEN = None


try:
    from cerebras.cloud.sdk._types import NOT_GIVEN as CEREBRAS_NOT_GIVEN
except ImportError:
    CEREBRAS_NOT_GIVEN = None


class DisplayNameFuncError(ValueError): ...


def print_call_link(call: "Call") -> None:
    if settings.should_print_call_link():
        print(f"{TRACE_CALL_EMOJI} {call.ui_url}")


FinishCallbackType = Callable[[Any, Optional[BaseException]], None]
OnOutputHandlerType = Callable[[Any, FinishCallbackType, Dict], Any]


def value_is_sentinel(param: Any) -> bool:
    return (
        param.default is None
        or param.default is OPENAI_NOT_GIVEN
        or param.default is COHERE_NOT_GIVEN
        or param.default is ANTHROPIC_NOT_GIVEN
        or param.default is MISTRAL_NOT_GIVEN
        or param.default is CEREBRAS_NOT_GIVEN
        or param.default is Ellipsis
    )


def _apply_fn_defaults_to_inputs(
    fn: typing.Callable, inputs: Mapping[str, typing.Any]
) -> dict[str, typing.Any]:
    inputs = {**inputs}
    sig = inspect.signature(fn)
    for param_name, param in sig.parameters.items():
        if param_name not in inputs:
            if param.default != inspect.Parameter.empty and not value_is_sentinel(
                param
            ):
                inputs[param_name] = param.default
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                inputs[param_name] = tuple()
            elif param.kind == inspect.Parameter.VAR_KEYWORD:
                inputs[param_name] = dict()
    return inputs


@runtime_checkable
class Op(Protocol):
    """
    The interface for Op-ified functions and methods.

    Op was previously a class, and has been converted to a Protocol to allow
    functions to pass for Op.  This is needed because many popular packages are
    using the `inspect` module for control flow, and Op instances don't always
    pass those checks.  In particular, `inspect.iscoroutinefunction` always
    fails for classes, even ones that implement async methods or protocols.

    Some of the attributes are carry-overs from when Op was a class.  We should
    consider removing the unnecessary ones where possible.
    - resolve_fn (I think you can just use the func itself?)
    - signature (just inspect the func)
    - _set_on_output_handler (does this have to be on the op?)
    - _on_output_handler (does this have to be on the op?)
    """

    name: str
    call_display_name: Union[str, Callable[["Call"], str]]
    signature: inspect.Signature
    ref: Optional[ObjectRef]
    resolve_fn: Callable

    postprocess_inputs: Optional[Callable[[dict[str, Any]], dict[str, Any]]]
    postprocess_output: Optional[Callable[..., Any]]

    call: Callable[..., Any]
    calls: Callable[..., "CallsIter"]

    # not sure if this is the best place for this, but kept for compat
    _set_on_output_handler: Callable[[OnOutputHandlerType], None]
    _on_output_handler: Optional[OnOutputHandlerType]

    __call__: Callable[..., Any]
    __self__: Any

    # `_tracing_enabled` is a runtime-only flag that can be used to disable
    # call tracing for an op. It is not persisted as a property of the op, but is
    # respected by the current execution context. It is an underscore property
    # because it is not intended to be used by users directly, but rather assists
    # with internal Weave behavior. If we find a need to expose this to users, we
    # should consider a more user-friendly API (perhaps a setter/getter) & whether
    # it disables child ops as well.
    _tracing_enabled: bool


def _set_on_output_handler(func: Op, on_output: OnOutputHandlerType) -> None:
    if func._on_output_handler is not None:
        raise ValueError("Cannot set on_output_handler multiple times")
    func._on_output_handler = on_output


def _is_unbound_method(func: Callable) -> bool:
    """Check if a function is a function defined on a class (an "unbound" method)

    In python3, the "unbound" method is just a function, but that distinction is
    not enough for our decorator because it needs to operate on both regular funcs
    and unbound methods at the same time.

    This check clarifies that distinction between function vs. unbound method.
    """
    sig = inspect.signature(func)
    params = list(sig.parameters.values())
    is_method = params and params[0].name in {"self", "cls"}

    return bool(is_method)


def _create_call(func: Op, *args: Any, **kwargs: Any) -> "Call":
    client = weave_client_context.require_weave_client()

    try:
        inputs = func.signature.bind(*args, **kwargs).arguments
    except TypeError as e:
        raise OpCallError(f"Error calling {func.name}: {e}")
    inputs_with_defaults = _apply_fn_defaults_to_inputs(func, inputs)

    # This should probably be configurable, but for now we redact the api_key
    if "api_key" in inputs_with_defaults:
        inputs_with_defaults["api_key"] = "REDACTED"

    # If/When we do memoization, this would be a good spot

    parent_call = call_context.get_current_call()
    attributes = call_attributes.get()

    return client.create_call(
        func,
        inputs_with_defaults,
        parent_call,
        display_name=func.call_display_name,
        attributes=attributes,
    )


def _execute_call(
    __op: Op,
    call: Any,
    *args: Any,
    __should_raise: bool = True,
    **kwargs: Any,
) -> Any:
    func = __op.resolve_fn
    client = weave_client_context.require_weave_client()
    has_finished = False

    def finish(output: Any = None, exception: Optional[BaseException] = None) -> None:
        nonlocal has_finished
        if has_finished:
            raise ValueError("Should not call finish more than once")

        client.finish_call(
            call,
            output,
            exception,
            postprocess_output=__op.postprocess_output,
        )
        if not call_context.get_current_call():
            print_call_link(call)

    def on_output(output: Any) -> Any:
        if handler := getattr(__op, "_on_output_handler", None):
            return handler(output, finish, call.inputs)
        finish(output)
        return output

    def process(res: Any) -> Any:
        res = box.box(res)
        try:
            # Here we do a try/catch because we don't want to
            # break the user process if we trip up on processing
            # the output
            res = on_output(res)
        except Exception as e:
            log_once(logger.error, f"Error capturing call output: {e}")
            if get_raise_on_captured_errors():
                raise
        finally:
            # Is there a better place for this? We want to ensure that even
            # if the final output fails to be captured, we still pop the call
            # so we don't put future calls under the old call.
            call_context.pop_call(call.id)

        return res, call

    def handle_exception(e: Exception) -> Any:
        finish(exception=e)
        if __should_raise:
            raise
        return None, call

    if inspect.iscoroutinefunction(func):

        async def _call_async() -> Coroutine[Any, Any, Any]:
            try:
                res = await func(*args, **kwargs)
            except Exception as e:
                return handle_exception(e)
            else:
                return process(res)

        return _call_async()

    try:
        res = func(*args, **kwargs)
    except Exception as e:
        handle_exception(e)
    else:
        return process(res)

    return None, call


def call(op: Op, *args: Any, **kwargs: Any) -> tuple[Any, "Call"]:
    """
    Executes the op and returns both the result and a Call representing the execution.

    This function will never raise.  Any errors are captured in the Call object.

    This method is automatically bound to any function decorated with `@weave.op`,
    allowing for usage like:

    ```python
    @weave.op
    def add(a: int, b: int) -> int:
        return a + b

    result, call = add.call(1, 2)
    ```
    """
    c = _create_call(op, *args, **kwargs)
    return _execute_call(op, c, *args, __should_raise=False, **kwargs)


def calls(op: Op) -> "CallsIter":
    """
    Get an iterator over all calls to this op.

    This method is automatically bound to any function decorated with `@weave.op`,
    allowing for usage like:

    ```python
    @weave.op
    def add(a: int, b: int) -> int:
        return a + b

    calls = add.calls()
    for call in calls:
        print(call)
    ```
    """
    client = weave_client_context.require_weave_client()
    return client._op_calls(op)


# Legacy decos
@overload
def op(name: Any) -> Any: ...
@overload
def op(input_type: Any, output_type: Any) -> Any: ...
@overload
def op(name: Any, input_type: Any, output_type: Any) -> Any: ...
@overload
def op(name: Any, output_type: Any) -> Any: ...
@overload
def op(name: Any, input_type: Any, output_type: Any, render_info: Any) -> Any: ...
@overload
def op(name: Any, input_type: Any, output_type: Any, pure: Any) -> Any: ...
@overload
def op(name: Any, input_type: Any, output_type: Any, hidden: Any) -> Any: ...
@overload
def op(
    input_type: Any = None,
    output_type: Any = None,
    refine_output_type: Any = None,
    name: Any = None,
    setter: Any = None,
    render_info: Any = None,
    hidden: Any = None,
    pure: Any = None,
    _op_def_class: Any = None,
    plugins: Any = None,
    mutation: Any = None,
    weavify: Any = None,
) -> Any: ...


# Modern decos
@overload
def op(func: Any) -> Op: ...


@overload
def op(
    *,
    call_display_name: Union[str, Callable[["Call"], str]],
) -> Callable[[Any], Op]:
    """Use call_display_name to set the display name of the traced call.

    When set as a callable, the callable must take in a Call object
    (which can have attributes like op_name, trace_id, etc.) and return
    the string to be used as the display name of the traced call."""
    ...


# type ignore here is because we have the legacy decorators above.  Once they are
# removed, we can remove the overloads this type ignore.
@overload
def op(*, name: str) -> Callable[[Any], Op]:  # type: ignore
    """Use name to set the name of the op itself."""
    ...


@overload
def op(
    *,
    postprocess_inputs: Callable[[dict[str, Any]], dict[str, Any]],
    postprocess_output: Callable[..., Any],
) -> Any:
    """
    Modify the inputs and outputs of an op before sending data to weave.

    This does not modify inputs or outputs at function call time, only when
    the data is sent to weave.
    """
    ...


def op(*args: Any, **kwargs: Any) -> Union[Callable[[Any], Op], Op]:
    """
    A decorator to weave op-ify a function or method.  Works for both sync and async.

    Decorated functions and methods can be called as normal, but will also
    automatically track calls in the Weave UI.

    If you don't call `weave.init` then the function will behave as if it were
    not decorated.


    Example usage:

    ```python
    import weave
    weave.init("my-project")

    @weave.op
    async def extract():
        return await client.chat.completions.create(
            model="gpt-4-turbo",
            messages=[
                {"role": "user", "content": "Create a user as JSON"},
            ],
        )

    await extract()  # calls the function and tracks the call in the Weave UI
    ```
    """
    if context_state.get_loading_built_ins():
        from weave.legacy.weave.decorator_op import op as legacy_op

        return legacy_op(*args, **kwargs)  # type: ignore

    def op_deco(func: Callable) -> Op:
        # Check function type
        sig = inspect.signature(func)
        is_method = _is_unbound_method(func)
        is_async = inspect.iscoroutinefunction(func)

        def create_wrapper(func: Callable) -> Op:
            if is_async:

                @wraps(func)
                async def wrapper(*args: Any, **kwargs: Any) -> Any:
                    if settings.should_disable_weave():
                        return await func(*args, **kwargs)
                    if weave_client_context.get_weave_client() is None:
                        return await func(*args, **kwargs)
                    if not wrapper._tracing_enabled:  # type: ignore
                        return await func(*args, **kwargs)
                    call = _create_call(wrapper, *args, **kwargs)  # type: ignore
                    res, _ = await _execute_call(wrapper, call, *args, **kwargs)  # type: ignore
                    return res
            else:

                @wraps(func)
                def wrapper(*args: Any, **kwargs: Any) -> Any:
                    if settings.should_disable_weave():
                        return func(*args, **kwargs)
                    if weave_client_context.get_weave_client() is None:
                        return func(*args, **kwargs)
                    if not wrapper._tracing_enabled:  # type: ignore
                        return func(*args, **kwargs)
                    call = _create_call(wrapper, *args, **kwargs)  # type: ignore
                    res, _ = _execute_call(wrapper, call, *args, **kwargs)  # type: ignore
                    return res

            # Tack these helpers on to our wrapper
            wrapper.resolve_fn = func  # type: ignore

            inferred_name = func.__qualname__ if is_method else func.__name__

            # funcs and methods defined inside another func will have the
            # name prefixed with {outer}.<locals>.{func_name}
            # this is noisy for us, so we strip it out
            inferred_name = inferred_name.split(".<locals>.")[-1]

            wrapper.name = kwargs.get("name", inferred_name)  # type: ignore
            wrapper.signature = sig  # type: ignore
            wrapper.ref = None  # type: ignore

            wrapper.postprocess_inputs = kwargs.get("postprocess_inputs")  # type: ignore
            wrapper.postprocess_output = kwargs.get("postprocess_output")  # type: ignore

            wrapper.call = partial(call, wrapper)  # type: ignore
            wrapper.calls = partial(calls, wrapper)  # type: ignore

            wrapper.__call__ = wrapper  # type: ignore
            wrapper.__self__ = wrapper  # type: ignore

            wrapper._set_on_output_handler = partial(_set_on_output_handler, wrapper)  # type: ignore
            wrapper._on_output_handler = None  # type: ignore

            wrapper._tracing_enabled = True  # type: ignore

            if callable(call_name_func := kwargs.get("call_display_name", "")):
                params = inspect.signature(call_name_func).parameters
                if len(params) != 1:
                    raise DisplayNameFuncError(
                        "`call_display_name` function must take exactly 1 argument (the Call object)"
                    )
            wrapper.call_display_name = call_name_func  # type: ignore

            return cast(Op, wrapper)

        return create_wrapper(func)

    if len(args) == 1 and len(kwargs) == 0 and callable(func := args[0]):
        return op_deco(func)

    return op_deco


def maybe_bind_method(func: Callable, self: Any = None) -> Union[Callable, MethodType]:
    """Bind a function to any object (even if it's not a class)

    If self is None, return the function as is.
    """
    if (sig := inspect.signature(func)) and sig.parameters.get("self"):
        if inspect.ismethod(func) and id(func.__self__) != id(self):
            raise ValueError("Cannot re-bind a method to an new object")
        return MethodType(func, self)
    return func


def maybe_unbind_method(oplike: Union[Op, MethodType, partial]) -> Op:
    """Unbind an Op-like method or partial to a plain Op function.

    For:
    - methods, remove set `self` param
    - partials, remove any preset params
    """
    if isinstance(oplike, MethodType):
        op = oplike.__func__
    elif isinstance(oplike, partial):  # Handle cases op is defined as
        op = oplike.func
    else:
        op = oplike

    return cast(Op, op)


def is_op(obj: Any) -> bool:
    if sys.version_info < (3, 12):
        return isinstance(obj, Op)

    return all(hasattr(obj, attr) for attr in Op.__annotations__)


def as_op(fn: Callable) -> Op:
    """Given a @weave.op() decorated function, return its Op.

    @weave.op() decorated functions are instances of Op already, so this
    function should be a no-op at runtime. But you can use it to satisfy type checkers
    if you need to access OpDef attributes in a typesafe way.

    Args:
        fn: A weave.op() decorated function.

    Returns:
        The Op of the function.
    """
    if not is_op(fn):
        raise ValueError("fn must be a weave.op() decorated function")

    return cast(Op, fn)


__docspec__ = [call, calls]



import dataclasses

import numpy as np
import pytest

from weave.legacy.weave import storage
from weave.legacy.weave import weave_types as types


def test_nparray():
    arr = np.array([[1, 2], [3, 4]])
    ref = storage.save(arr)
    arr2 = ref.get()
    assert (arr == arr2).all()
    arr3 = storage.get(str(ref))
    assert (arr == arr3).all()


def test_list_int():
    arr = [1, 2, 3]
    ref = storage.save(arr)
    arr2 = ref.get()
    assert arr[0] == arr2[0]
    assert arr[1] == arr2[1]
    assert arr[2] == arr2[2]


def test_list_dict():
    arr = [{"a": 5, "b": 3}, {"a": 6, "b": 7}]
    ref = storage.save(arr)
    arr2 = ref.get()
    assert arr[0]["a"] == arr2[0]["a"]
    assert arr[0]["b"] == arr2[0]["b"]
    assert arr[1]["a"] == arr2[1]["a"]
    assert arr[1]["b"] == arr2[1]["b"]


# def test_list_dict_with_maybe():
#     arr = [{'a': 5, 'b': 3}, {'a': 6, 'c': 7}]
#     ref = storage.save(arr)
#     arr2 = ref.get()
#     assert arr[0]['a'] == arr2[0]['a']
#     assert arr[0]['b'] == arr2[0]['b']
#     assert arr[1]['a'] == arr2[1]['a']
#     assert arr[1]['c'] == arr2[1]['c']


def test_list_nested_dict():
    arr = [{"a": 5, "b": {"c": 9}}, {"a": 6, "b": {"c": 10}}]
    ref = storage.save(arr)
    arr2 = ref.get()
    # TODO: make object equality work
    assert arr2[0]["a"] == arr[0]["a"]
    assert arr2[0]["b"] == arr[0]["b"]
    assert arr2[1]["a"] == arr[1]["a"]
    assert arr2[1]["b"] == arr[1]["b"]


def test_list_double_nested_dict():
    arr = [{"a": 5, "b": {"c": {"d": 11}}}, {"a": 6, "b": {"c": {"d": 10}}}]
    ref = storage.save(arr)
    arr2 = ref.get()
    # TODO: make object equality work
    assert arr2[0]["a"] == arr[0]["a"]
    assert arr2[0]["b"] == arr[0]["b"]
    assert arr2[1]["a"] == arr[1]["a"]
    assert arr2[1]["b"] == arr[1]["b"]


class SomeObj:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y


class SomeObjType(types.ObjectType):
    instance_classes = SomeObj
    instance_class = SomeObj
    name = "someobj"

    def property_types(self):
        return {"x": types.Int(), "y": types.Int()}


def test_list_nested_obj():
    arr = [{"a": 5, "b": SomeObj(1, 2)}, {"a": 6, "b": SomeObj(3, 4)}]
    ref = storage.save(arr)
    arr2 = ref.get()
    assert isinstance(arr2[0]["b"], SomeObj)
    assert arr2[0]["a"] == arr[0]["a"]
    assert arr2[0]["b"] == arr[0]["b"]
    assert arr2[1]["a"] == arr[1]["a"]
    assert arr2[1]["b"] == arr[1]["b"]


class SomeGenericObj:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y


@dataclasses.dataclass(frozen=True)
class SomeGenericObjType(types.ObjectType):
    instance_classes = SomeGenericObj
    instance_class = SomeGenericObj
    name = "somegenericobj"

    x: types.Any

    def property_types(self):
        return {"x": self.x, "y": types.Int()}


def test_list_nested_generic_obj():
    arr = [{"a": 5, "b": SomeGenericObj(1, 2)}, {"a": 6, "b": SomeGenericObj(3, 4)}]
    ref = storage.save(arr)
    arr2 = ref.get()
    assert isinstance(arr2[0]["b"], SomeGenericObj)
    assert arr2[0]["a"] == arr[0]["a"]
    assert arr2[0]["b"] == arr[0]["b"]
    assert arr2[1]["a"] == arr[1]["a"]
    assert arr2[1]["b"] == arr[1]["b"]


def test_list_nested_generic_obj_deep():
    arr = [
        {"a": 5, "b": SomeGenericObj({"x": 14, "y": SomeObj(9, 99)}, -4)},
        {"a": 6, "b": SomeGenericObj({"x": 21, "y": SomeObj(100, 100)}, -44)},
    ]
    ref = storage.save(arr)
    arr2 = ref.get()
    assert isinstance(arr2[0]["b"], SomeGenericObj)
    assert arr2[0]["a"] == arr[0]["a"]
    assert arr2[0]["b"] == arr[0]["b"]
    assert arr2[1]["a"] == arr[1]["a"]
    assert arr2[1]["b"] == arr[1]["b"]


def test_list_nested_lists():
    arr = [
        {"a": 5, "b": [{"x": 4, "y": 1}, {"x": 5, "y": 2}]},
        {"a": 6, "b": [{"x": 8, "y": 2}, {"x": 9, "y": 3}]},
    ]
    ref = storage.save(arr)
    arr2 = ref.get()
    assert arr2[0]["b"][0]["x"] == arr[0]["b"][0]["x"]
    assert arr2[0]["b"][1]["y"] == arr[0]["b"][1]["y"]


def test_list_nested_lists_with_objs():
    arr = [
        {"a": 5, "b": [{"x": 4, "y": SomeObj(9, 99)}, {"x": 5, "y": SomeObj(14, 13)}]},
        {
            "a": 6,
            "b": [{"x": 8, "y": SomeObj(10, 101)}, {"x": 9, "y": SomeObj(12, 11)}],
        },
    ]
    ref = storage.save(arr)
    arr2 = ref.get()
    assert arr2[0]["b"][0]["x"] == arr[0]["b"][0]["x"]
    assert arr2[0]["b"][1]["y"] == arr[0]["b"][1]["y"]


@pytest.mark.skip("Need union support in mappers_arrow")
def test_list_nested_lists_with_objs_1missing():
    arr = [
        {"a": 5, "b": [{"x": 4, "y": SomeObj(9, 99)}, {"x": 5, "y": SomeObj(14, 13)}]},
        {"a": 6, "b": [{"x": 8}, {"x": 9, "y": SomeObj(12, 11)}]},
    ]
    ref = storage.save(arr)
    arr2 = ref.get()
    assert arr2[0]["b"][0]["x"] == arr[0]["b"][0]["x"]
    assert arr2[0]["b"][1]["y"] == arr[0]["b"][1]["y"]
    assert arr2[1]["b"][0]["x"] == arr[1]["b"][0]["x"]
    assert arr2[1]["b"][1]["x"] == arr[1]["b"][1]["x"]

    # TODO: this is broken!
    # assert "y" not in arr2[1]["b"][0]

    assert arr2[1]["b"][1]["y"] == arr[1]["b"][1]["y"]


def test_serializers():
    arr = [{"a": 5, "b": np.ones((2, 3))}, {"a": 7, "b": np.ones((2, 3))}]
    ref = storage.save(arr)
    arr2 = ref.get()
    assert arr2[0]["a"] == arr[0]["a"]
    assert (arr2[0]["b"] == arr[0]["b"]).all()



import os
import typing

from weave.legacy.weave import artifact_local, graph, runs

from . import ref_base
from . import errors
from weave.legacy.weave import weave_types as types


def get_obj_creator(ref: ref_base.Ref) -> typing.Optional[runs.Run]:
    # ref.backend.references(ref.artifact, "Run", {"dir": "input"})

    # creator = ref.backend.get_creator(ref.artifact)
    # # OR
    # backend = ref.backend.filter(type="Run", referenced=ref.artifact)
    # Extremely inefficient!
    # TODO
    for art_name in os.listdir(artifact_local.local_artifact_dir()):
        if (
            art_name.startswith("run-")
            and not art_name.endswith("-output")
            and artifact_local.local_artifact_exists(art_name, "latest")
        ):
            try:
                cache_obj = artifact_local.get_local_version(art_name, "latest")
            except errors.WeaveSerializeError:
                # This happens because we don't load all ecosystem modules, and so we can't
                # deserializes everything.
                continue

            def _is_creator(run):
                if isinstance(run.output, ref_base.Ref) and str(run.output) == str(ref):
                    # If any input is also the ref, this run did not create obj, since
                    # the obj already existed. This fixes an infinite loop where list-createIndexCheckpointTag
                    # which just returns its input would be treated as a the obj creator
                    # TODO: This whole thing is a pile of hacks, not production ready! Fix! We should
                    #     not need heuristics.
                    # TODO: for one, if we order all the artifacts by created_at, then
                    #     the first one will be the creator. Can't do that in this branch
                    #     since it doesn't have the created at change.
                    if any(str(input) == str(ref) for input in run.inputs.values()):
                        return False
                    return True

            if isinstance(cache_obj._ref.type, types.List):
                for run in cache_obj:
                    if _is_creator(run):
                        return run
            else:
                if _is_creator(cache_obj):
                    return cache_obj
    return None


def get_obj_expr(obj):
    obj_type = types.TypeRegistry.type_of(obj)
    if not isinstance(obj, ref_base.Ref):
        return graph.ConstNode(obj_type, obj)
    run = get_obj_creator(obj)
    if run is None:
        return graph.ConstNode(obj_type, obj)
    return graph.OutputNode(
        obj.type,
        run.op_name,
        {k: get_obj_expr(input) for k, input in run.inputs.items()},
    )


def used_by(ref, op_name: str) -> list[runs.Run]:
    users = []
    for artifact_name in os.listdir(artifact_local.local_artifact_dir()):
        if artifact_name.startswith("run-") and not artifact_name.endswith("-output"):
            run = artifact_local.get_local_version(artifact_name, "latest")
            if run is None:
                continue
            run_inputs = list(run.inputs.values())
            if run.op_name == op_name and run_inputs:
                input0 = run_inputs[0]
                if (
                    isinstance(input0, artifact_local.LocalArtifactRef)
                    and ref.uri == input0.uri
                ):
                    users.append(run)
    return users



import datetime
import os
import time

from weave.legacy.weave import cache, environment


def test_lru_time_window_cache():
    curtime = {"t": datetime.datetime(2020, 1, 1)}

    def now_fn():
        # Time ticks forward 1s at each step
        t = curtime["t"]
        curtime["t"] = t + datetime.timedelta(seconds=1)
        return t

    c = cache.LruTimeWindowCache(datetime.timedelta(seconds=5), now_fn=now_fn)
    c.set("foo", "a")
    c.set("boink", "will_prune")
    c.set("foo", "b")
    c.set("bar", "d")
    c.set("baz", "e")
    assert list(c._cache.items()) == [
        ((None, "boink"), (datetime.datetime(2020, 1, 1, 0, 0, 1), "will_prune")),
        ((None, "foo"), (datetime.datetime(2020, 1, 1, 0, 0, 2), "b")),
        ((None, "bar"), (datetime.datetime(2020, 1, 1, 0, 0, 3), "d")),
        ((None, "baz"), (datetime.datetime(2020, 1, 1, 0, 0, 4), "e")),
    ]

    c.get("foo")
    c.set("bozo", "y")
    c.set("bar", "f")
    c.set("bar2", "h")
    assert list(c._cache.items()) == [
        ((None, "baz"), (datetime.datetime(2020, 1, 1, 0, 0, 4), "e")),
        ((None, "foo"), (datetime.datetime(2020, 1, 1, 0, 0, 5), "b")),
        ((None, "bozo"), (datetime.datetime(2020, 1, 1, 0, 0, 6), "y")),
        ((None, "bar"), (datetime.datetime(2020, 1, 1, 0, 0, 7), "f")),
        ((None, "bar2"), (datetime.datetime(2020, 1, 1, 0, 0, 8), "h")),
    ]


def test_bucket_timestamp():
    day_in_seconds = 60 * 60 * 24
    cacheTimestamps = []
    for i in range(10):
        cacheTimestamps.append(int(cache.bucket_timestamp(i), 10))

    # Assert that the 0 is always 0
    assert cacheTimestamps[0] == 0

    # Assert that the timestamps are bucketed into the correct intervals
    for i in range(1, 10):
        assert cacheTimestamps[i] % day_in_seconds == 0
        assert cacheTimestamps[i] % (day_in_seconds * i) == 0

    # Assert that running the same function should bucket into the same interval
    for i in range(10):
        assert cacheTimestamps[i] == int(cache.bucket_timestamp(i), 10)


def test_clear_cache():
    # Test that the clear_cache function works
    day_in_seconds = 60 * 60 * 24

    # ensure cache directory exists
    cache_dir = environment.weave_filesystem_dir()
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    # check cache directory file count
    orig_file_count = len(os.listdir(cache_dir))
    paths = []

    # create 10 cache directories
    # 5 in the past, 5 in the future
    for i in [-10, -9, -8, -7, -6, 1, 2, 3, 4, 5]:
        path = os.path.join(cache_dir, str(int(time.time() + (i * day_in_seconds))))
        paths.append(path)
        os.makedirs(path)

    # check cache directory file count is correct
    assert len(os.listdir(cache_dir)) == orig_file_count + len(paths)

    # clear the cache
    # should delete 5 past directories
    cache.clear_cache()

    # check cache directory file count is correct
    assert len(os.listdir(cache_dir)) == orig_file_count + len(paths) / 2

    # delete all created driectories
    for path in paths:
        if os.path.exists(path):
            os.rmdir(path)

    # Ensure cache directory is in the same state as before the test
    assert len(os.listdir(cache_dir)) == orig_file_count



import typing

import pytest

import weave
from weave.legacy.weave import panels, storage
from weave.legacy.weave.artifact_fs import BranchPointType, FilesystemArtifactRef
from weave.legacy.weave.artifact_local import (
    LocalArtifact,
    LocalArtifactRef,
    WeaveLocalArtifactURI,
)
from weave.legacy.weave.artifact_wandb import (
    WandbArtifact,
    WandbArtifactRef,
    WeaveWBArtifactURI,
    likely_commit_hash,
)
from weave.legacy.weave.uris import WeaveURI


def test_publish_values(user_by_api_key_in_env):
    data = ["a", "b", "c"]
    ref = storage.publish(data, "weave/list")
    assert ref.get() == data


def test_publish_panel(user_by_api_key_in_env):
    from weave.legacy.weave import panel_util

    table_obj = panels.Table(
        panel_util.make_node(
            [
                {"a": 1, "b": 2, "c": 3},
                {"a": 4, "b": 5, "c": 6},
                {"a": 7, "b": 8, "c": 9},
            ]
        ),
        columns=[
            lambda row: row["a"],
            lambda row: row["b"],
            lambda row: row["c"],
        ],
    )
    ref = storage.publish(table_obj, "weave/table")
    assert isinstance(ref, WandbArtifactRef)


def test_publish_table(user_by_api_key_in_env):
    table_obj = panels.Table(
        weave.save(
            [
                {"a": 1, "b": 2, "c": 3},
                {"a": 4, "b": 5, "c": 6},
                {"a": 7, "b": 8, "c": 9},
            ]
        ),
        columns=[
            lambda row: row["a"],
            lambda row: row["b"],
            lambda row: row["c"],
        ],
    )
    ref = storage.publish(table_obj, "weave/table")
    assert (
        ref.get().input_node.from_op.inputs["uri"].val.startswith("wandb-artifact://")
    )


def test_publish_group(user_by_api_key_in_env):
    local_data = weave.save(
        [
            {"a": 1, "b": 2, "c": 3},
            {"a": 4, "b": 5, "c": 6},
            {"a": 7, "b": 8, "c": 9},
        ]
    )

    group = panels.Group(
        items={
            "table": panels.Table(
                local_data,
                columns=[
                    lambda row: row["a"],
                    lambda row: row["b"] * 2,
                ],
            ),
            "plot": lambda table: panels.Plot(
                table.all_rows(),
                x=lambda row: row["c_0"],
                y=lambda row: row["c_1"],
            ),
        }
    )

    res = storage.publish(group, "weave/group")


"""
TODO: These tests were written in prep for the launch and need to be cleaned up
and made more maintainable/understandable. This is a followup action on Tim.

The following tests exercise the save & publish flows. There are 2 dimensions to test:

* The `Location` of the data.
* The `Action` to take on the data.

There are 6 possible `Location` types:

1. Remote with commit hash (eg. `wandb-artifact://ENTITY/PROJECT/NAME:HASH`)
2. Remote with branch (eg. `wandb-artifact://ENTITY/PROJECT/NAME:BRANCH`)
3-4. Local with commit hash (eg. `wandb-artifact://ENTITY/PROJECT/NAME:HASH`). Variants:
    a. No branchpoint
    b. With branchpoint
5-6. Local with branch (eg. `wandb-artifact://ENTITY/PROJECT/NAME:BRANCH`). Variants:
    a. No branchpoint
    b. With branchpoint

There are 3 actions:

1. `Persist`: Directly persist data (eq. save or publish directly)
2. `Mutate`: Generate new version via Mutation
3. `Merge`: Apply changes to branchpoint (only applicable to locations with branchpoint)
   * The branchpoint URI can be any of the 6 location types above.

In summary, there will be 22 tests:

* 6 `Persist` tests
* 6 `Mutate` tests
* 2 * 6 `Merge` tests (2 types of branchpoints, and 6 branch location types.)
   * -2 cases that are not possible
"""


# Persist tests
def test_persist_to_remote_with_commit_hash(user_by_api_key_in_env):
    data = ["test_data"]
    target_entity = user_by_api_key_in_env.username
    target_project_name = "test_project"
    target_artifact_name = "test_artifact"
    expected_commit_hash = "74d5ba98aca469b59e18"
    branch_name = None

    p_ref = storage._direct_publish(
        obj=data,
        name=target_artifact_name,
        wb_project_name=target_project_name,
        wb_entity_name=target_entity,
        branch_name=branch_name,
    )

    _perform_post_persist_assertions(
        is_local=False,
        data=data,
        p_ref=p_ref,
        target_artifact_name=target_artifact_name,
        expected_commit_hash=expected_commit_hash,
        target_entity=target_entity,
        target_project_name=target_project_name,
        branch_name=branch_name,
    )


def test_persist_to_remote_with_branch(user_by_api_key_in_env):
    data = ["test_data"]
    target_entity = user_by_api_key_in_env.username
    target_project_name = "test_project"
    target_artifact_name = "test_artifact"
    expected_commit_hash = "74d5ba98aca469b59e18"
    branch_name = "test_branch"

    p_ref = storage._direct_publish(
        obj=data,
        name=target_artifact_name,
        wb_project_name=target_project_name,
        wb_entity_name=target_entity,
        branch_name=branch_name,
    )

    _perform_post_persist_assertions(
        is_local=False,
        data=data,
        p_ref=p_ref,
        target_artifact_name=target_artifact_name,
        expected_commit_hash=expected_commit_hash,
        target_entity=target_entity,
        target_project_name=target_project_name,
        branch_name=branch_name,
    )


def test_persist_to_local_with_commit_hash(user_by_api_key_in_env):
    data = ["test_data"]
    target_artifact_name = "test_artifact"
    expected_commit_hash = "b11179315e19b4207282"
    branch_name = None

    p_ref = storage.direct_save(
        obj=data,
        name=target_artifact_name,
        branch_name=branch_name,
    )

    _perform_post_persist_assertions(
        is_local=True,
        data=data,
        p_ref=p_ref,
        target_artifact_name=target_artifact_name,
        expected_commit_hash=expected_commit_hash,
        branch_name=branch_name,
    )


# Skipping as this one does not make logical sense.
# def test_persist_to_local_with_commit_hash_and_branchpoint(user_by_api_key_in_env):
#     pass


def test_persist_to_local_with_branch(user_by_api_key_in_env):
    data = ["test_data"]
    target_artifact_name = "test_artifact"
    expected_commit_hash = "b11179315e19b4207282"
    branch_name = "my_branch"

    p_ref = storage.direct_save(
        obj=data,
        name=target_artifact_name,
        branch_name=branch_name,
    )

    _perform_post_persist_assertions(
        is_local=True,
        data=data,
        p_ref=p_ref,
        target_artifact_name=target_artifact_name,
        expected_commit_hash=expected_commit_hash,
        branch_name=branch_name,
    )


def test_persist_to_local_with_branch_and_branchpoint(user_by_api_key_in_env):
    data = ["test_data"]
    target_artifact_name = "test_artifact"
    expected_commit_hash = "b11179315e19b4207282"
    branch_name = "my_branch"

    source_branch_name = "initial_branch"
    source_commit_hash = "741eb73e40d3b38b046b"

    # First, save a local artifact with a branch
    p_ref = storage.direct_save(
        obj=["initial_data"],
        name=target_artifact_name,
        branch_name=source_branch_name,
    )

    # Then, save a new local artifact with a branchpoint
    new_p_ref = storage.direct_save(
        obj=data,
        name=target_artifact_name,
        branch_name=branch_name,
        source_branch_name=source_branch_name,
    )

    _perform_post_persist_assertions(
        is_local=True,
        data=data,
        p_ref=new_p_ref,
        target_artifact_name=target_artifact_name,
        expected_commit_hash=expected_commit_hash,
        branch_name=branch_name,
        expected_branchpoint=BranchPointType(
            commit=source_commit_hash,
            n_commits=1,
            original_uri=str(p_ref.branch_uri).replace("/obj", ""),
        ),
    )


def _perform_post_persist_assertions(
    is_local: bool,
    data: typing.Any,
    p_ref: FilesystemArtifactRef,
    target_artifact_name: str,
    expected_commit_hash: str,
    target_entity: typing.Optional[str] = None,
    target_project_name: typing.Optional[str] = None,
    branch_name: typing.Optional[str] = None,
    expected_branchpoint: typing.Optional[BranchPointType] = None,
):
    p_uri: WeaveURI
    p_art: typing.Union[WandbArtifact, LocalArtifact]

    if is_local:
        assert isinstance(p_ref, LocalArtifactRef)
        p_art = p_ref.artifact
        assert isinstance(p_art, LocalArtifact)
        p_uri = p_art.uri_obj
        assert isinstance(p_uri, WeaveLocalArtifactURI)
        assert target_entity is None
        assert target_project_name is None
        expected_uri_with_hash = (
            f"local-artifact:///{target_artifact_name}:{expected_commit_hash}/obj"
        )
        expected_uri_with_branch = None
        if branch_name:
            expected_uri_with_branch = (
                f"local-artifact:///{target_artifact_name}:{branch_name}/obj"
            )
    else:
        assert isinstance(p_ref, WandbArtifactRef)
        p_art = p_ref.artifact
        assert isinstance(p_art, WandbArtifact)
        p_uri = p_art.uri_obj
        assert isinstance(p_uri, WeaveWBArtifactURI)
        assert target_entity is not None
        assert target_project_name is not None
        expected_uri_with_hash = f"wandb-artifact:///{target_entity}/{target_project_name}/{target_artifact_name}:{expected_commit_hash}/obj"
        expected_uri_with_branch = None
        if branch_name:
            expected_uri_with_branch = f"wandb-artifact:///{target_entity}/{target_project_name}/{target_artifact_name}:{branch_name}/obj"

    # Data Assertions
    get_op = weave.legacy.weave.ops.get(expected_uri_with_hash)
    get_res = weave.use(get_op)
    assert p_ref.obj == get_res == data
    # Only check branch if expected_uri_with_branch is not None
    if expected_uri_with_branch:
        get_op = weave.legacy.weave.ops.get(expected_uri_with_branch)
        get_res = weave.use(get_op)
        assert get_res == data

    # Identity Assertions (entity, project, name, version)
    if not is_local:
        assert isinstance(p_uri, WeaveWBArtifactURI)
        assert p_uri.entity_name == target_entity
        assert p_uri.project_name == target_project_name
    assert p_ref.name == p_art.name == p_uri.name == target_artifact_name
    assert likely_commit_hash(expected_commit_hash) and (
        p_ref.version == p_art.version == p_uri.version == expected_commit_hash
    )
    if not is_local:
        assert isinstance(p_art, WandbArtifact)
        assert p_art.commit_hash == expected_commit_hash
    else:
        assert p_art.version == expected_commit_hash

    assert p_ref.branch == p_art.branch
    assert p_art.branch == branch_name

    # Branchpoint Assertions
    assert p_ref.branch_point == p_art.branch_point == expected_branchpoint

    # URI Assertions
    assert p_ref.initial_uri == expected_uri_with_branch or expected_uri_with_hash
    assert p_ref.uri == expected_uri_with_hash
    assert p_ref.branch_uri == expected_uri_with_branch or expected_uri_with_hash


# Mutate tests
def test_mutate_remote_with_commit_hash(user_by_api_key_in_env):
    data = ["test_data"]
    target_entity = user_by_api_key_in_env.username
    target_project_name = "test_project"
    target_artifact_name = "test_artifact"
    expected_commit_hash = "74d5ba98aca469b59e18"
    branch_name = None

    new_data = ["test_data_2"]
    new_commit_hash = "61f78c8877df22942d23"

    p_ref = storage._direct_publish(
        obj=data,
        name=target_artifact_name,
        wb_project_name=target_project_name,
        wb_entity_name=target_entity,
        branch_name=branch_name,
    )

    assert p_ref.version == expected_commit_hash

    obj = weave.legacy.weave.ops.get(p_ref.uri)[0]
    obj.set(new_data[0])
    new_branch_name = f"user-{p_ref.version}"
    new_uri = f"local-artifact:///{target_artifact_name}:{new_branch_name}/obj"
    new_p_ref = LocalArtifactRef.from_str(new_uri)

    _perform_post_persist_assertions(
        is_local=True,
        data=new_data,
        p_ref=new_p_ref,
        target_artifact_name=target_artifact_name,
        expected_commit_hash=new_commit_hash,
        branch_name=new_branch_name,
        expected_branchpoint=BranchPointType(
            commit=expected_commit_hash,
            n_commits=1,
            original_uri=str(p_ref.branch_uri).replace("/obj", ""),
        ),
    )


def test_mutate_remote_with_branch(user_by_api_key_in_env):
    data = ["test_data"]
    target_entity = user_by_api_key_in_env.username
    target_project_name = "test_project"
    target_artifact_name = "test_artifact"
    expected_commit_hash = "74d5ba98aca469b59e18"
    branch_name = "remote_branch"

    new_data = ["test_data_2"]
    new_commit_hash = "61f78c8877df22942d23"

    p_ref = storage._direct_publish(
        obj=data,
        name=target_artifact_name,
        wb_project_name=target_project_name,
        wb_entity_name=target_entity,
        branch_name=branch_name,
    )

    assert p_ref.version == expected_commit_hash
    assert branch_name in p_ref.branch_uri

    obj = weave.legacy.weave.ops.get(p_ref.branch_uri)[0]
    obj.set(new_data[0])
    new_branch_name = f"user-{branch_name}"
    new_p_ref = LocalArtifactRef.from_str(
        f"local-artifact:///{target_artifact_name}:{new_branch_name}/obj"
    )

    _perform_post_persist_assertions(
        is_local=True,
        data=new_data,
        p_ref=new_p_ref,
        target_artifact_name=target_artifact_name,
        expected_commit_hash=new_commit_hash,
        branch_name=new_branch_name,
        expected_branchpoint=BranchPointType(
            commit=expected_commit_hash,
            n_commits=1,
            original_uri=str(p_ref.branch_uri).replace("/obj", ""),
        ),
    )


def test_mutate_local_with_commit_hash(user_by_api_key_in_env, new_branch_name=None):
    data = ["test_data"]
    target_artifact_name = "test_artifact"
    expected_commit_hash = "b11179315e19b4207282"
    branch_name = None

    new_data = ["test_data_2"]
    new_commit_hash = "61f78c8877df22942d23"

    p_ref = storage.direct_save(
        obj=data,
        name=target_artifact_name,
        branch_name=branch_name,
    )

    assert p_ref.version == expected_commit_hash

    obj = weave.legacy.weave.ops.get(p_ref.uri)[0]
    if new_branch_name == None:
        obj.set(new_data[0])
        new_branch_name = new_commit_hash
        expected_branchpoint = None
    else:
        obj.set(new_data[0], root_args={"branch": new_branch_name})
        branch_name = new_branch_name
        expected_branchpoint = BranchPointType(
            commit=expected_commit_hash,
            n_commits=1,
            original_uri=str(p_ref.branch_uri).replace("/obj", ""),
        )
    new_p_ref = LocalArtifactRef.from_str(
        f"local-artifact:///{target_artifact_name}:{new_branch_name}/obj"
    )

    _perform_post_persist_assertions(
        is_local=True,
        data=new_data,
        p_ref=new_p_ref,
        target_artifact_name=target_artifact_name,
        expected_commit_hash=new_commit_hash,
        branch_name=branch_name,
        expected_branchpoint=expected_branchpoint,
    )


def test_mutate_local_with_commit_hash_and_branchpoint(
    user_by_api_key_in_env, new_branch_name=None
):
    data = ["test_data"]
    target_entity = user_by_api_key_in_env.username
    target_project_name = "test_project"
    target_artifact_name = "test_artifact"
    expected_commit_hash = "74d5ba98aca469b59e18"
    branch_name = "remote_branch"

    new_data = ["test_data_2"]
    new_commit_hash = "61f78c8877df22942d23"

    p_ref = storage._direct_publish(
        obj=data,
        name=target_artifact_name,
        wb_project_name=target_project_name,
        wb_entity_name=target_entity,
        branch_name=branch_name,
    )

    assert p_ref.version == expected_commit_hash
    assert branch_name in p_ref.branch_uri

    obj = weave.legacy.weave.ops.get(p_ref.branch_uri)[0]
    obj.set("PLACEHOLDER")
    mod_ref = LocalArtifactRef.from_uri(
        WeaveLocalArtifactURI.parse(
            f"local-artifact:///{target_artifact_name}:user-{branch_name}/obj"
        )
    )
    obj = weave.legacy.weave.ops.get(mod_ref.branch_uri)[0]
    if new_branch_name == None:
        obj.set(new_data[0])
        branch_name = None
        new_branch_name = new_commit_hash
        expected_branchpoint = BranchPointType(
            commit=expected_commit_hash,
            n_commits=2,
            original_uri=str(p_ref.branch_uri).replace("/obj", ""),
        )
    else:
        obj.set(new_data[0], root_args={"branch": new_branch_name})
        branch_name = new_branch_name
        expected_branchpoint = BranchPointType(
            commit="0dc63f69254034abd6f9",
            n_commits=1,
            original_uri=str(mod_ref.branch_uri).replace("/obj", ""),
        )

    new_p_ref = LocalArtifactRef.from_str(
        f"local-artifact:///{target_artifact_name}:{new_branch_name}/obj"
    )

    _perform_post_persist_assertions(
        is_local=True,
        data=new_data,
        p_ref=new_p_ref,
        target_artifact_name=target_artifact_name,
        expected_commit_hash=new_commit_hash,
        branch_name=branch_name,
        expected_branchpoint=expected_branchpoint,
    )


def test_mutate_local_with_branch(user_by_api_key_in_env, new_branch_name=None):
    data = ["test_data"]
    target_artifact_name = "test_artifact"
    expected_commit_hash = "b11179315e19b4207282"
    branch_name = "my_branch"

    new_data = ["test_data_2"]
    new_commit_hash = "61f78c8877df22942d23"

    p_ref = storage.direct_save(
        obj=data,
        name=target_artifact_name,
        branch_name=branch_name,
    )

    assert p_ref.version == expected_commit_hash

    obj = weave.legacy.weave.ops.get(p_ref.branch_uri)[0]
    if new_branch_name == None:
        obj.set(new_data[0])
        new_branch_name = new_commit_hash
        expected_branchpoint = None
    else:
        obj.set(new_data[0], root_args={"branch": new_branch_name})
        branch_name = new_branch_name
        expected_branchpoint = BranchPointType(
            commit=expected_commit_hash,
            n_commits=1,
            original_uri=str(p_ref.branch_uri).replace("/obj", ""),
        )
    new_branch_name = branch_name
    new_p_ref = LocalArtifactRef.from_str(
        f"local-artifact:///{target_artifact_name}:{new_branch_name}/obj"
    )

    _perform_post_persist_assertions(
        is_local=True,
        data=new_data,
        p_ref=new_p_ref,
        target_artifact_name=target_artifact_name,
        expected_commit_hash=new_commit_hash,
        branch_name=new_branch_name,
        expected_branchpoint=expected_branchpoint,
    )


def test_mutate_local_with_branch_and_branchpoint(user_by_api_key_in_env):
    data = ["test_data"]
    target_entity = user_by_api_key_in_env.username
    target_project_name = "test_project"
    target_artifact_name = "test_artifact"
    expected_commit_hash = "74d5ba98aca469b59e18"
    branch_name = "remote_branch"

    new_data = ["test_data_2"]
    new_commit_hash = "61f78c8877df22942d23"

    p_ref = storage._direct_publish(
        obj=data,
        name=target_artifact_name,
        wb_project_name=target_project_name,
        wb_entity_name=target_entity,
        branch_name=branch_name,
    )

    assert p_ref.version == expected_commit_hash
    assert branch_name in p_ref.branch_uri

    obj = weave.legacy.weave.ops.get(p_ref.branch_uri)[0]
    obj.set("PLACEHOLDER")
    obj = weave.legacy.weave.ops.get(
        f"local-artifact:///{target_artifact_name}:user-{branch_name}/obj"
    )[0]
    obj.set(new_data[0])
    new_p_ref = LocalArtifactRef.from_str(
        f"local-artifact:///{target_artifact_name}:user-{branch_name}/obj"
    )

    _perform_post_persist_assertions(
        is_local=True,
        data=new_data,
        p_ref=new_p_ref,
        target_artifact_name=target_artifact_name,
        expected_commit_hash=new_commit_hash,
        branch_name=f"user-{branch_name}",
        expected_branchpoint=BranchPointType(
            commit=expected_commit_hash,
            n_commits=2,
            original_uri=str(p_ref.branch_uri).replace("/obj", ""),
        ),
    )


# Merge Tests


def test_merge_from_local_with_commit_hash_onto_remote_with_commit_hash(
    user_by_api_key_in_env,
):
    test_mutate_remote_with_commit_hash(user_by_api_key_in_env)

    merged_uri = weave.legacy.weave.ops.merge_artifact(
        weave.legacy.weave.ops.get("local-artifact:///test_artifact:61f78c8877df22942d23/obj")
    )

    assert merged_uri.startswith("wandb-artifact://")
    assert "5a61da7b34feb72c2af8" in merged_uri

    new_p_ref = WandbArtifactRef.from_str(merged_uri)

    _perform_post_persist_assertions(
        is_local=False,
        target_entity=user_by_api_key_in_env.username,
        target_project_name="test_project",
        data=["test_data_2"],
        p_ref=new_p_ref,
        target_artifact_name="test_artifact",
        expected_commit_hash="5a61da7b34feb72c2af8",
        branch_name=None,
    )


def test_merge_from_local_with_commit_hash_onto_remote_with_branch(
    user_by_api_key_in_env,
):
    test_mutate_remote_with_branch(user_by_api_key_in_env)

    merged_uri = weave.legacy.weave.ops.merge_artifact(
        weave.legacy.weave.ops.get("local-artifact:///test_artifact:61f78c8877df22942d23/obj")
    )

    assert merged_uri.startswith("wandb-artifact://")
    assert "remote_branch" in merged_uri

    new_p_ref = WandbArtifactRef.from_str(merged_uri)

    _perform_post_persist_assertions(
        is_local=False,
        target_entity=user_by_api_key_in_env.username,
        target_project_name="test_project",
        data=["test_data_2"],
        p_ref=new_p_ref,
        target_artifact_name="test_artifact",
        expected_commit_hash="5a61da7b34feb72c2af8",
        branch_name="remote_branch",
    )


def test_merge_from_local_with_commit_hash_onto_local_with_commit_hash(
    user_by_api_key_in_env,
):
    test_mutate_local_with_commit_hash(user_by_api_key_in_env, "new_branch_name")

    merged_uri = weave.legacy.weave.ops.merge_artifact(
        weave.legacy.weave.ops.get("local-artifact:///test_artifact:61f78c8877df22942d23/obj")
    )

    assert merged_uri.startswith("local-artifact://")
    assert "61f78c8877df22942d23" in merged_uri

    new_p_ref = WandbArtifactRef.from_str(merged_uri)

    _perform_post_persist_assertions(
        is_local=True,
        data=["test_data_2"],
        p_ref=new_p_ref,
        target_artifact_name="test_artifact",
        expected_commit_hash="61f78c8877df22942d23",
        branch_name=None,
    )


# def test_merge_from_local_with_commit_hash_onto_local_with_commit_hash_and_branchpoint(
#     user_by_api_key_in_env,
# ):
#     # Leaving the test stub here for completeness of the permutations, but I am
#     # not convinced it is even possible to get into this state.
#     pass


def test_merge_from_local_with_commit_hash_onto_local_with_branch(
    user_by_api_key_in_env,
):
    test_mutate_local_with_branch(user_by_api_key_in_env, "new_branch_name")

    merged_uri = weave.legacy.weave.ops.merge_artifact(
        weave.legacy.weave.ops.get("local-artifact:///test_artifact:61f78c8877df22942d23/obj")
    )

    assert merged_uri.startswith("local-artifact://")
    assert "my_branch" in merged_uri

    new_p_ref = WandbArtifactRef.from_str(merged_uri)

    _perform_post_persist_assertions(
        is_local=True,
        data=["test_data_2"],
        p_ref=new_p_ref,
        target_artifact_name="test_artifact",
        expected_commit_hash="61f78c8877df22942d23",
        branch_name="my_branch",
    )


def test_merge_from_local_with_commit_hash_onto_local_with_branch_and_branchpoint(
    user_by_api_key_in_env,
):
    test_mutate_local_with_commit_hash_and_branchpoint(
        user_by_api_key_in_env, "new_branch_name"
    )

    merged_uri = weave.legacy.weave.ops.merge_artifact(
        weave.legacy.weave.ops.get("local-artifact:///test_artifact:61f78c8877df22942d23/obj")
    )

    assert merged_uri.startswith("local-artifact://")
    assert "remote_branch" in merged_uri

    new_p_ref = WandbArtifactRef.from_str(merged_uri)

    _perform_post_persist_assertions(
        is_local=True,
        data=["test_data_2"],
        p_ref=new_p_ref,
        target_artifact_name="test_artifact",
        expected_commit_hash="61f78c8877df22942d23",
        branch_name="user-remote_branch",
        expected_branchpoint=BranchPointType(
            n_commits=2,
            commit="74d5ba98aca469b59e18",
            original_uri=f"wandb-artifact:///{user_by_api_key_in_env.username}/test_project/test_artifact:remote_branch",
        ),
    )

    merged_uri_2 = weave.legacy.weave.ops.merge_artifact(weave.legacy.weave.ops.get(merged_uri))
    new_p_ref_2 = WandbArtifactRef.from_str(merged_uri_2)

    assert merged_uri_2.startswith("wandb-artifact://")
    assert "remote_branch" in merged_uri_2

    new_p_ref = WandbArtifactRef.from_str(merged_uri_2)

    _perform_post_persist_assertions(
        is_local=False,
        data=["test_data_2"],
        p_ref=new_p_ref_2,
        target_artifact_name="test_artifact",
        expected_commit_hash="5a61da7b34feb72c2af8",
        branch_name="remote_branch",
        target_entity=user_by_api_key_in_env.username,
        target_project_name="test_project",
    )


# Merge Tests


def test_merge_from_local_with_branch_onto_remote_with_commit_hash(
    user_by_api_key_in_env,
):
    test_mutate_remote_with_commit_hash(user_by_api_key_in_env)

    merged_uri = weave.legacy.weave.ops.merge_artifact(
        weave.legacy.weave.ops.get(
            "local-artifact:///test_artifact:user-74d5ba98aca469b59e18/obj"
        )
    )

    assert merged_uri.startswith("wandb-artifact://")
    assert "5a61da7b34feb72c2af8" in merged_uri

    new_p_ref = WandbArtifactRef.from_str(merged_uri)

    _perform_post_persist_assertions(
        is_local=False,
        target_entity=user_by_api_key_in_env.username,
        target_project_name="test_project",
        data=["test_data_2"],
        p_ref=new_p_ref,
        target_artifact_name="test_artifact",
        expected_commit_hash="5a61da7b34feb72c2af8",
        branch_name=None,
    )


def test_merge_from_local_with_branch_onto_remote_with_branch(user_by_api_key_in_env):
    test_mutate_remote_with_branch(user_by_api_key_in_env)

    merged_uri = weave.legacy.weave.ops.merge_artifact(
        weave.legacy.weave.ops.get("local-artifact:///test_artifact:user-remote_branch/obj")
    )

    assert merged_uri.startswith("wandb-artifact://")
    assert "remote_branch" in merged_uri

    new_p_ref = WandbArtifactRef.from_str(merged_uri)

    _perform_post_persist_assertions(
        is_local=False,
        target_entity=user_by_api_key_in_env.username,
        target_project_name="test_project",
        data=["test_data_2"],
        p_ref=new_p_ref,
        target_artifact_name="test_artifact",
        expected_commit_hash="5a61da7b34feb72c2af8",
        branch_name="remote_branch",
    )


def test_merge_from_local_with_branch_onto_local_with_commit_hash(
    user_by_api_key_in_env,
):
    test_mutate_local_with_commit_hash(user_by_api_key_in_env, "new_branch_name")

    merged_uri = weave.legacy.weave.ops.merge_artifact(
        weave.legacy.weave.ops.get("local-artifact:///test_artifact:new_branch_name/obj")
    )

    assert merged_uri.startswith("local-artifact://")
    assert "61f78c8877df22942d23" in merged_uri

    new_p_ref = WandbArtifactRef.from_str(merged_uri)

    _perform_post_persist_assertions(
        is_local=True,
        data=["test_data_2"],
        p_ref=new_p_ref,
        target_artifact_name="test_artifact",
        expected_commit_hash="61f78c8877df22942d23",
        branch_name=None,
    )


# def test_merge_from_local_with_branch_onto_local_with_commit_hash_and_branchpoint(
#     user_by_api_key_in_env,
# ):
#     # Leaving the test stub here for completeness of the permutations, but I am
#     # not convinced it is even possible to get into this state.
#     pass


def test_merge_from_local_with_branch_onto_local_with_branch(user_by_api_key_in_env):
    test_mutate_local_with_branch(user_by_api_key_in_env, "new_branch_name")

    merged_uri = weave.legacy.weave.ops.merge_artifact(
        weave.legacy.weave.ops.get("local-artifact:///test_artifact:new_branch_name/obj")
    )

    assert merged_uri.startswith("local-artifact://")
    assert "my_branch" in merged_uri

    new_p_ref = WandbArtifactRef.from_str(merged_uri)

    _perform_post_persist_assertions(
        is_local=True,
        data=["test_data_2"],
        p_ref=new_p_ref,
        target_artifact_name="test_artifact",
        expected_commit_hash="61f78c8877df22942d23",
        branch_name="my_branch",
    )


def test_merge_from_local_with_branch_onto_local_with_branch_and_branchpoint(
    user_by_api_key_in_env,
):
    test_mutate_local_with_commit_hash_and_branchpoint(
        user_by_api_key_in_env, "new_branch_name"
    )

    merged_uri = weave.legacy.weave.ops.merge_artifact(
        weave.legacy.weave.ops.get("local-artifact:///test_artifact:new_branch_name/obj")
    )

    assert merged_uri.startswith("local-artifact://")
    assert "remote_branch" in merged_uri

    new_p_ref = WandbArtifactRef.from_str(merged_uri)

    _perform_post_persist_assertions(
        is_local=True,
        data=["test_data_2"],
        p_ref=new_p_ref,
        target_artifact_name="test_artifact",
        expected_commit_hash="61f78c8877df22942d23",
        branch_name="user-remote_branch",
        expected_branchpoint=BranchPointType(
            n_commits=2,
            commit="74d5ba98aca469b59e18",
            original_uri=f"wandb-artifact:///{user_by_api_key_in_env.username}/test_project/test_artifact:remote_branch",
        ),
    )

    merged_uri_2 = weave.legacy.weave.ops.merge_artifact(weave.legacy.weave.ops.get(merged_uri))
    new_p_ref_2 = WandbArtifactRef.from_str(merged_uri_2)

    assert merged_uri_2.startswith("wandb-artifact://")
    assert "remote_branch" in merged_uri_2

    new_p_ref = WandbArtifactRef.from_str(merged_uri_2)

    _perform_post_persist_assertions(
        is_local=False,
        data=["test_data_2"],
        p_ref=new_p_ref_2,
        target_artifact_name="test_artifact",
        expected_commit_hash="5a61da7b34feb72c2af8",
        branch_name="remote_branch",
        target_entity=user_by_api_key_in_env.username,
        target_project_name="test_project",
    )


## Custom Flow


def test_publish_saved_node(user_by_api_key_in_env):
    data = ["test_publish_saved_node"]
    saved = weave.save(data)
    assert saved.from_op.inputs["uri"].val.startswith("local-artifact://")
    assert weave.use(saved) == data

    published_art_uri = weave.legacy.weave.ops.publish_artifact(saved, "my_list", None, None)
    assert published_art_uri.startswith("wandb-artifact://")
    assert weave.use(weave.get(published_art_uri)) == data


def _uri_is_local(uri: str) -> bool:
    return uri.startswith("local-artifact://")


def _uri_is_remote(uri: str) -> bool:
    return uri.startswith("wandb-artifact://")


def _get_uri_version(uri: str) -> str:
    return uri.split(":")[-1].split("/")[0]


def _replace_uri_version(uri: str, new_version: str) -> str:
    return uri.replace(_get_uri_version(uri), new_version)


def _get_uri_from_get_node(node: weave.legacy.weave.graph.Node) -> str:
    return node.from_op.inputs["uri"].val  # type: ignore


def test_end_to_end_save_and_publish_flow(user_by_api_key_in_env):
    # Step 1: Save some data locally.
    data = ["data_a"]
    saved_node = weave.save(data)
    saved_uri = _get_uri_from_get_node(saved_node)
    saved_version = _get_uri_version(saved_uri)
    assert _uri_is_local(saved_uri)
    assert weave.use(saved_node) == data

    # Step 2: Make a branching mutation.
    weave.legacy.weave.ops.set(
        saved_node[0],
        "test_publish_saved_node_execution_2",
        root_args={"branch": "my-branch"},
    )
    branched_data = ["test_publish_saved_node_execution_2"]
    branched_uri = saved_uri.replace(saved_version, "my-branch")
    branched_node = weave.legacy.weave.ops.get(branched_uri)
    assert weave.use(branched_node) == branched_data

    # Step 3: Merge the change back into the main branch.
    merged_uri = weave.legacy.weave.ops.merge_artifact(branched_node)
    assert merged_uri.startswith("local-artifact://")
    assert merged_uri != saved_uri != branched_uri
    merged_node = weave.legacy.weave.ops.get(merged_uri)
    assert weave.use(merged_node) == branched_data

    # Step 4: Publish the new version remotely
    published_uri = weave.legacy.weave.ops.publish_artifact(
        merged_node, "my_list", None, None
    )
    assert published_uri.startswith("wandb-artifact://")
    assert weave.use(weave.get(published_uri)) == branched_data

    # Step 5: Modify the remote version
    weave.legacy.weave.ops.set(
        weave.legacy.weave.ops.get(published_uri)[0],
        "test_publish_saved_node_execution_3",
        root_args={"branch": "my-branch-2"},
    )
    published_branched_data = ["test_publish_saved_node_execution_3"]
    published_branched_uri = "local-artifact:///my_list:my-branch-2/obj"
    published_branched_node = weave.legacy.weave.ops.get(published_branched_uri)
    assert weave.use(published_branched_node) == published_branched_data

    # Step 6: Merge the remote change back into the main branch.
    published_merged_uri = weave.legacy.weave.ops.merge_artifact(published_branched_node)
    assert published_merged_uri.startswith("wandb-artifact://")
    assert published_merged_uri != published_uri
    published_merged_node = weave.legacy.weave.ops.get(published_merged_uri)
    assert weave.use(published_merged_node) == published_branched_data



# Tests for the AWL interface

import pytest

import weave
import weave.legacy.weave.weave_types as types
from weave.legacy.weave import box, ops_arrow
from weave.legacy.weave.arrow.convert import to_arrow
from weave.legacy.weave.language_features.tagging import tag_store, tagged_value_type

from weave.legacy.tests.util.concrete_tagged_value import (
    TaggedValue,
    concrete_from_tagstore,
    concrete_to_tagstore,
)


def to_awl(l: list) -> ops_arrow.ArrowWeaveList:
    tagged = concrete_to_tagstore(l)
    return ops_arrow.to_arrow(tagged)


def from_awl(awl: ops_arrow.ArrowWeaveList) -> list:
    tagged = awl.to_pylist_tagged()
    return concrete_from_tagstore(tagged)


def test_column():
    awl = to_awl(
        [
            TaggedValue({"a": 5}, {"b": 6, "c": 7}),
            TaggedValue({"a": 9}, {"b": 10, "c": 12}),
        ]
    )
    col_b = awl.column("b")
    assert col_b.object_type == tagged_value_type.TaggedValueType(
        weave.types.TypedDict({"a": weave.types.Int()}), weave.types.Int()
    )
    expected = [
        TaggedValue({"a": 5}, 6),
        TaggedValue({"a": 9}, 10),
    ]
    assert from_awl(col_b) == expected


def test_apply():
    awl = to_awl(
        [
            TaggedValue({"a": 5}, {"b": 6, "c": 7}),
            TaggedValue({"a": 9}, {"b": 10, "c": 12}),
        ]
    )
    res = awl.apply(lambda row: (row["b"] + row["c"]) * 2)
    assert res.object_type == tagged_value_type.TaggedValueType(
        weave.types.TypedDict({"a": weave.types.Int()}), weave.types.Number()
    )
    assert from_awl(res) == [
        TaggedValue({"a": 5}, 26),
        TaggedValue({"a": 9}, 44),
    ]


def test_join2():
    l1 = to_awl([{"a": 5, "b": 6}, {"a": 7, "b": 8}])
    l2 = to_awl([{"a": 5, "b": 9}, {"a": 10, "b": 14}])
    res = l1.join2(
        l2, lambda row: row["a"], lambda row: row["a"], leftOuter=True, rightOuter=True
    )
    assert from_awl(res) == [
        TaggedValue({"joinObj": 5}, {"a1": {"a": 5, "b": 6}, "a2": {"a": 5, "b": 9}}),
        TaggedValue({"joinObj": 7}, {"a1": {"a": 7, "b": 8}, "a2": None}),
        TaggedValue(tag={"joinObj": 10}, value={"a1": None, "a2": {"a": 10, "b": 14}}),
    ]


@pytest.mark.skip()
def test_join2_different_types():
    l1 = to_awl([{"a": {"x": 5}, "b": 6}, {"a": {"x": 5}, "b": 8}])
    l2 = to_awl([{"a": {"x": 6, "j": 9}, "b": 9}, {"a": {"x": 6, "y": 7}, "b": 14}])
    res = l1.join2(
        l2, lambda row: row["a"], lambda row: row["a"], leftOuter=True, rightOuter=True
    )
    assert from_awl(res) == [
        TaggedValue({"joinObj": 5}, {"a1": {"a": 5, "b": 6}, "a2": {"a": 5, "b": 9}}),
        TaggedValue({"joinObj": 7}, {"a1": {"a": 7, "b": 8}, "a2": None}),
        TaggedValue(tag={"joinObj": 10}, value={"a1": None, "a2": {"a": 10, "b": 14}}),
    ]


def test_tagged_awl_sum():
    data = [[1.4, 0], [3.4, 0], [None, 5]]
    for i, item in enumerate(data):
        item[0] = tag_store.add_tags(box.box(item[0]), {"mytag": f"{i}-abcd"})

    tagtype = tagged_value_type.TaggedValueType(
        types.TypedDict(property_types={"mytag": types.String()}),
        types.UnionType(types.NoneType(), types.Number()),
    )
    awl = to_arrow(
        data,
        wb_type=types.List(
            object_type=types.List(object_type=types.UnionType(tagtype, types.Int()))
        ),
    )

    node = weave.save(awl)
    value = weave.use(node.listnumbersum())

    assert value.to_pylist_notags() == [1.4, 3.4, 5]



# Note: `uuid_utils` is a pure UUID drop in replacement that leverages Rust's
# UUID library for performance. We may want to write our own UUIDv7
# implementation to support python < 3.8. However, Weave already requires >=3.9,
# so we can just use the library.
import uuid_utils as uuid


def generate_id() -> str:
    """Should be used to generate IDs for trace calls.
    We use UUIDv7, which has a timestamp prefix, so that the IDs, while random,
    are sortable by time. See RFC9562 https://www.rfc-editor.org/rfc/rfc9562

    Random space is 2^74, which is less than 2^122 (UUIDv4), but still plenty
    for our use case.
    """
    return str(uuid.uuid7())
